#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Integration tests for constraint scaling by expected magnitude in Custom Scaling.

Tests will use the Gibbs Reactor example as the test case as it is very
poorly scaled in native form.

Author: Andrew Lee
"""

import os
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    Constraint,
    Suffix,
    TransformationFactory,
    value,
    Var,
    units,
)

from idaes.core import FlowsheetBlock
from idaes.models.unit_models.gibbs_reactor import GibbsReactor
from idaes.models.properties.activity_coeff_models.methane_combustion_ideal import (
    MethaneParameterBlock as MethaneCombustionParameterBlock,
)
from idaes.core.util import to_json, from_json, StoreSpec
from idaes.core.util.scaling import jacobian_cond
from idaes.core.scaling import AutoScaler, CustomScalerBase, set_scaling_factor


FILENAME = "gibbs_solution.json"
local_path = os.path.dirname(os.path.realpath(__file__))
fname = os.path.join(local_path, FILENAME)


def build_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = MethaneCombustionParameterBlock()

    m.fs.unit = GibbsReactor(
        property_package=m.fs.properties,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    m.fs.unit.inlet.flow_mol[0].fix(230.0)
    m.fs.unit.inlet.mole_frac_comp[0, "H2"].fix(0.0435)
    m.fs.unit.inlet.mole_frac_comp[0, "N2"].fix(0.6522)
    m.fs.unit.inlet.mole_frac_comp[0, "O2"].fix(0.1739)
    m.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(1e-5)
    m.fs.unit.inlet.mole_frac_comp[0, "CH4"].fix(0.1304)
    m.fs.unit.inlet.mole_frac_comp[0, "CO"].fix(1e-5)
    m.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1e-5)
    m.fs.unit.inlet.mole_frac_comp[0, "NH3"].fix(1e-5)
    m.fs.unit.inlet.temperature[0].fix(1500.0)
    m.fs.unit.inlet.pressure[0].fix(101325.0)

    m.fs.unit.outlet.temperature[0].fix(2844.38)
    m.fs.unit.deltaP.fix(0)

    return m


@pytest.fixture
def gibbs():
    m = build_model()

    # Load solution
    from_json(m, fname=fname, wts=StoreSpec().value())
    # Make sure we have no suffixes loaded
    assert not hasattr(m.fs.unit, "scaling_factor")

    # Autoscale variables by magnitude
    scaler = AutoScaler()
    scaler.scale_variables_by_magnitude(m)
    assert hasattr(m.fs.unit, "scaling_factor")

    return m


@pytest.mark.integration
def test_verify_model_load(gibbs):
    for v in gibbs.component_data_objects(ctype=Var, descend_into=True):
        assert v.value is not None


@pytest.mark.integration
def test_autoscale_L2_norm(gibbs):
    scaler = AutoScaler()
    scaler.scale_constraints_by_jacobian_norm(gibbs, norm=2)

    scaled = jacobian_cond(gibbs, scaled=True)

    assert scaled == pytest.approx(2510.945, rel=1e-5)


@pytest.mark.integration
def test_autoscale_L1_norm(gibbs):
    scaler = AutoScaler()
    scaler.scale_constraints_by_jacobian_norm(gibbs, norm=1)

    scaled = jacobian_cond(gibbs, scaled=True)

    assert scaled == pytest.approx(2986.994, rel=1e-5)


@pytest.mark.integration
def test_nominal_magnitude_harmonic(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_value(c, scheme="harmonic_mean")

    assert jacobian_cond(gibbs, scaled=True) == pytest.approx(2.83944e12, rel=1e-5)


@pytest.mark.integration
def test_nominal_magnitude_inv_max(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_value(c, scheme="inverse_maximum")

    assert jacobian_cond(gibbs, scaled=True) == pytest.approx(784576, rel=1e-5)


@pytest.mark.integration
def test_nominal_magnitude_inv_min(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_value(c, scheme="inverse_minimum")

    assert jacobian_cond(gibbs, scaled=True) == pytest.approx(5.601e12, rel=1e-5)


@pytest.mark.integration
def test_nominal_magnitude_inv_sum(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_value(c, scheme="inverse_sum")

    assert jacobian_cond(gibbs, scaled=True) == pytest.approx(1501632, rel=1e-5)


@pytest.mark.integration
def test_nominal_magnitude_inv_rss(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_value(c, scheme="inverse_root_sum_squared")

    assert jacobian_cond(gibbs, scaled=True) == pytest.approx(959994, rel=1e-5)


@pytest.mark.integration
def test_scale_constraint_by_nominal_derivative_2norm_perfect_information(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_derivative_norm(c)

    scaled = jacobian_cond(gibbs, scaled=True)
    assert scaled == pytest.approx(3.07419e06, rel=1e-5)


@pytest.mark.integration
def test_scale_constraint_by_nominal_derivative_2norm_imperfect_information():
    # Build a fresh model with no scaling factors
    model = build_model()

    # Set imperfect scaling factors for all variables, representing an initial "best-guess"
    # Feed states are known exactly - set scaling based on these
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].flow_mol, 1 / 230
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].flow_mol_phase, 1 / 230
    )  # Only 1 phase, so we "know" this
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2"], 1 / 0.0435
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["N2"], 1 / 0.6522
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["O2"], 1 / 0.1739
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO2"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CH4"],
        1 / 0.1304,
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2O"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["NH3"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].temperature, 1 / 1500
    )
    set_scaling_factor(model.fs.unit.control_volume.properties_in[0.0].pressure, 1e-5)
    # Assume user does not know anything about enthalpy

    # Best guesses for unit model and outlet state conditions
    set_scaling_factor(model.fs.unit.control_volume.heat[0.0], 1e-6)

    set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].flow_mol, 1e-2)
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase, 1e-2
    )  # Only 1 phase, so we "know" this
    # N2 is inert, so will be order 0.1, assume CH4 and H2 are near-totally consumed, assume most O2 consumed
    # Assume moderate amounts of CO2 and H2O, small amounts of CO, trace NH3 NH3
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["N2"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["O2"], 1e2
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO2"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CH4"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO"], 1e3
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2O"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["NH3"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].temperature, 1e-3
    )
    set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5)

    scaler = CustomScalerBase()
    for c in model.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_derivative_norm(c)

    scaled = jacobian_cond(model, scaled=True)
    assert scaled == pytest.approx(1.06128e11, rel=1e-5)


@pytest.mark.integration
def test_scale_constraint_by_nominal_derivative_1norm_perfect_information(gibbs):
    scaler = CustomScalerBase()

    for c in gibbs.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_derivative_norm(c, norm=1)

    scaled = jacobian_cond(gibbs, scaled=True)
    assert scaled == pytest.approx(2.060153e06, rel=1e-5)


@pytest.mark.integration
def test_scale_constraint_by_nominal_derivative_clean_up(gibbs):
    # Confirm that the scaler did not change any values in the model
    # Build a fresh model with no scaling factors or initial values
    model = build_model()
    # Clone the model to use as a reference case
    refmodel = model.clone()

    # Set imperfect scaling factors for all variables, representing an initial "best-guess"
    # Feed states are known exactly - set scaling based on these
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].flow_mol, 1 / 230
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].flow_mol_phase, 1 / 230
    )  # Only 1 phase, so we "know" this
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2"], 1 / 0.0435
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["N2"], 1 / 0.6522
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["O2"], 1 / 0.1739
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO2"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CH4"],
        1 / 0.1304,
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["CO"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["H2O"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].mole_frac_comp["NH3"], 1e5
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_in[0.0].temperature, 1 / 1500
    )
    set_scaling_factor(model.fs.unit.control_volume.properties_in[0.0].pressure, 1e-5)
    # Assume user does not know anything about enthalpy

    # Best guesses for unit model and outlet state conditions
    set_scaling_factor(model.fs.unit.control_volume.heat[0.0], 1e-6)

    set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].flow_mol, 1e-2)
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase, 1e-2
    )  # Only 1 phase, so we "know" this
    # N2 is inert, so will be order 0.1, assume CH4 and H2 are near-totally consumed, assume most O2 consumed
    # Assume moderate amounts of CO2 and H2O, small amounts of CO, trace NH3
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["N2"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["O2"], 1e2
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO2"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CH4"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["CO"], 1e3
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["H2O"], 10
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp["NH3"], 1e4
    )
    set_scaling_factor(
        model.fs.unit.control_volume.properties_out[0.0].temperature, 1e-3
    )
    set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5)

    scaler = CustomScalerBase()
    for c in model.component_data_objects(ctype=Constraint, descend_into=True):
        scaler.scale_constraint_by_nominal_derivative_norm(c)

    # Iterate overall all variables in model and compare value to reference model
    for v in model.component_data_objects(ctype=Var, descend_into=True):
        refv = refmodel.find_component(v.name)

        assert v.value == refv.value


if __name__ == "__main__":
    # Run file to regenerate solution json
    m = build_model()

    initialization_tester(
        m,
        optarg={"tol": 1e-6},
        state_args={
            "temperature": 2844.38,
            "pressure": 101325.0,
            "flow_mol": 251.05,
            "mole_frac_comp": {
                "CH4": 1e-5,
                "CO": 0.0916,
                "CO2": 0.0281,
                "H2": 0.1155,
                "H2O": 0.1633,
                "N2": 0.5975,
                "NH3": 1e-5,
                "O2": 0.0067,
            },
        },
    )

    scaler = AutoScaler()
    scaler.scale_variables_by_magnitude(m)
    scaler.scale_constraints_by_jacobian_norm(m)

    solver = get_solver("ipopt_v2", writer_config={"scale_model": True})
    results = solver.solve(m, tee=True)

    # Check for optimal solution
    assert_optimal_termination(results)

    to_json(m, fname=fname, human_read=True, wts=StoreSpec().value())



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from copy import deepcopy
from math import sqrt

import numpy as np

from .unit_cell_lattice import UnitCell, UnitCellLattice
from ..geometry import Cube
from ..tiling import CubicTiling
from ..transform_func import ScaleFunc, RotateFunc
from ...util.util import ListHasPoint


class DiamondLattice(UnitCellLattice):
    RefIAD = sqrt(3) / 4

    # === STANDARD CONSTRUCTOR
    def __init__(self, IAD):
        RefUnitCellShape = Cube(1, BotBackLeftCorner=np.array([0, 0, 0], dtype=float))
        RefUnitCellTiling = CubicTiling(RefUnitCellShape)
        RefFracPositions = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.5, 0.5, 0.0]),
            np.array([0.0, 0.5, 0.5]),
            np.array([0.5, 0.0, 0.5]),
            np.array([0.25, 0.25, 0.25]),
            np.array([0.25, 0.75, 0.75]),
            np.array([0.75, 0.25, 0.75]),
            np.array([0.75, 0.75, 0.25]),
        ]
        RefUnitCell = UnitCell(RefUnitCellTiling, RefFracPositions)
        UnitCellLattice.__init__(self, RefUnitCell)
        self._IAD = (
            DiamondLattice.RefIAD
        )  # IAD is set correctly after calling applyTransF
        self.applyTransF(ScaleFunc(IAD / DiamondLattice.RefIAD))
        self._NthNeighbors = [
            [
                [
                    np.array([0.25, 0.25, 0.25]),
                    np.array([-0.25, -0.25, 0.25]),
                    np.array([-0.25, 0.25, -0.25]),
                    np.array([0.25, -0.25, -0.25]),
                ],
                [
                    np.array([-0.25, -0.25, -0.25]),
                    np.array([0.25, 0.25, -0.25]),
                    np.array([0.25, -0.25, 0.25]),
                    np.array([-0.25, 0.25, 0.25]),
                ],
            ],
            [
                [
                    np.array([0.0, 0.5, 0.5]),
                    np.array([0.0, 0.5, -0.5]),
                    np.array([0.0, -0.5, 0.5]),
                    np.array([0.0, -0.5, -0.5]),
                    np.array([0.5, 0.5, 0.0]),
                    np.array([0.5, 0.0, 0.5]),
                    np.array([0.5, -0.5, 0.0]),
                    np.array([0.5, 0.0, -0.5]),
                    np.array([-0.5, 0.5, 0.0]),
                    np.array([-0.5, 0.0, 0.5]),
                    np.array([-0.5, -0.5, 0.0]),
                    np.array([-0.5, 0.0, -0.5]),
                ],
                [
                    np.array([0.0, 0.5, 0.5]),
                    np.array([0.0, 0.5, -0.5]),
                    np.array([0.0, -0.5, 0.5]),
                    np.array([0.0, -0.5, -0.5]),
                    np.array([0.5, 0.5, 0.0]),
                    np.array([0.5, 0.0, 0.5]),
                    np.array([0.5, -0.5, 0.0]),
                    np.array([0.5, 0.0, -0.5]),
                    np.array([-0.5, 0.5, 0.0]),
                    np.array([-0.5, 0.0, 0.5]),
                    np.array([-0.5, -0.5, 0.0]),
                    np.array([-0.5, 0.0, -0.5]),
                ],
            ],
        ]
        self._typeDict = {0: 0, 3: 1}
        self._relativePositions = {
            0: np.array([0.0, 0.0, 0.0]),
            3: np.array([0.25, 0.25, 0.25]),
        }

    # === CONSTRUCTOR - Aligned with {100}
    @classmethod
    def alignedWith100(cls, IAD):
        return cls(IAD)  # Default implementation

    # === CONSTRUCTOR - Aligned with {110}
    @classmethod
    def alignedWith110(cls, IAD):
        result = cls(IAD)
        thetaX = 0
        thetaY = np.pi * 0.25
        thetaZ = 0
        result.applyTransF(RotateFunc.fromXYZAngles(thetaX, thetaY, thetaZ))
        return result

    # === CONSTRUCTOR - Aligned with {111}
    @classmethod
    def alignedWith111(cls, IAD, blnTrianglesAlignedWithX=True):
        result = cls(IAD)
        thetaX = -np.pi * 0.25
        thetaY = -np.arctan2(-sqrt(2), 2)
        thetaZ = np.pi * 0.5 if blnTrianglesAlignedWithX else 0
        result.applyTransF(RotateFunc.fromXYZAngles(thetaX, thetaY, thetaZ))
        return result

    # === CONSTRUCTOR - Aligned with {xyz}
    @classmethod
    def alignedWith(cls, IAD, MI):
        if (type(MI) is str) and (len(MI) == 3) and all(x.isdigit() for x in MI):
            if MI in ["100", "010", "001"]:
                return cls(IAD)
            elif MI in ["110", "101", "011"]:
                return cls.alignedWith110(IAD)
            elif MI == "111":
                return cls.alignedWith111(IAD)
            else:
                result = cls(IAD)
                a = np.array([0.0, 0.0, 1.0])
                b = np.array([float(MI[0]), float(MI[1]), float(MI[2])])
                axis = np.cross(a, b)
                angle = np.arccos(
                    np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))
                )
                result.applyTransF(RotateFunc.fromAxisAngle(axis, angle))
                return result
        return ValueError("DiamondLattice.alignedWith: Input direction is not correct.")

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._IAD *= TransF.Scale[0]
            else:
                raise ValueError(
                    "DiamondLattice.applyTransF: Can only scale isometrically"
                )
        UnitCellLattice.applyTransF(self, TransF)

    # === AUXILIARY METHODS
    def _getPointType(self, P):
        return (int(round(P[0] * 4)) + int(round(P[1] * 4)) + int(round(P[2] * 4))) % 4

    # === PROPERTY EVALUATION METHODS
    # NOTE: inherited from UnitCellLattice
    # def isOnLattice(self,P):

    def areNeighbors(self, P1, P2):
        return np.linalg.norm(P2 - P1) <= self.IAD

    def getNeighbors(self, P, layer=1):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        if PType not in self._typeDict.keys():
            raise ValueError("DiamondLattice.getNeighbors Should never reach here!")
        if layer > len(self._NthNeighbors):
            self._calculateNeighbors(layer)
        NBs = deepcopy(self._NthNeighbors[layer - 1][self._typeDict[PType]])
        for NeighP in NBs:
            NeighP += RefP
            self._convertFromReference(NeighP)
        return NBs

    def _calculateNeighbors(self, layer):
        NList = []
        for k, v in self._typeDict.items():
            tmp = [np.array([0, 0, 0], dtype=float)]
            for nb in self._NthNeighbors:
                tmp.extend(nb[v])
            NList.append(tmp)
        for _ in range(layer - len(self._NthNeighbors)):
            tmp = [[] for _ in self._typeDict.keys()]
            for k, v in self._typeDict.items():
                for P in self._NthNeighbors[len(self._NthNeighbors) - 1][v]:
                    PType = self._getPointType(P + self._relativePositions[k])
                    for Q in self._NthNeighbors[0][self._typeDict[PType]]:
                        N = P + Q
                        if not ListHasPoint(NList[v], N, 0.001 * DiamondLattice.RefIAD):
                            tmp[v].append(N)
                            NList[v].append(N)
            self._NthNeighbors.append(tmp)

    def isASite(self, P):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        return PType == 0

    def isBSite(self, P):
        RefP = self._getConvertToReference(P)
        PType = self._getPointType(RefP)
        return PType == 3

    def setDesign(self, D, AType, BType):
        for i, P in enumerate(D.Canvas.Points):
            if self.isASite(P):
                D.setContent(i, AType)
            elif self.isBSite(P):
                D.setContent(i, BType)
            else:
                raise ValueError("setDesign can not set site not on lattice")

    # === BASIC QUERY METHODS
    @property
    def IAD(self):
        return self._IAD

    @property
    def Diamond100LayerSpacing(self):
        return self.IAD / sqrt(3)

    @property
    def Diamond110LayerSpacing(self):
        return self.IAD * sqrt(2) / sqrt(3)

    @property
    def Diamond111LayerSpacing(self):
        return self.IAD * 4 / 3

    @property
    def Diamond112LayerSpacing(self):
        return self.IAD * sqrt(2) / 3

    def getLayerSpacing(self, MI):
        if (type(MI) is str) and (len(MI) == 3) and all(x.isdigit() for x in MI):
            if MI in ["100", "010", "001"]:
                return self.Diamond100LayerSpacing
            elif MI in ["110", "101", "011"]:
                return self.Diamond110LayerSpacing
            elif MI == "111":
                return self.Diamond111LayerSpacing
            elif MI in ["112", "121", "211"]:
                return self.Diamond112LayerSpacing
            else:
                raise NotImplementedError(
                    "DiamondLattice.getLayerSpacing: Input direction is not supported."
                )
        return ValueError(
            "DiamondLattice.getLayerSpacing: Input direction is not correct."
        )

    def getShellSpacing(self, MI):
        if (type(MI) is str) and (len(MI) == 3) and all(x.isdigit() for x in MI):
            if MI in ["100", "010", "001", "110", "101", "011", "111"]:
                return self.IAD * sqrt(8) / sqrt(3)
            elif MI in ["112", "121", "211"]:
                return self.IAD * sqrt(2) / sqrt(3)
            else:
                raise NotImplementedError(
                    "DiamondLattice.getShellSpacing: Input direction is not supported."
                )
        return ValueError("The input direction is not correct.")

    def getUniqueLayerCount(self, MI):
        if (type(MI) is str) and (len(MI) == 3) and all(x.isdigit() for x in MI):
            if MI in ["100", "010", "001"]:
                return 4
            elif MI in ["110", "101", "011"]:
                return 2
            elif MI == "111":
                return 3
            elif MI in ["112", "121", "211"]:
                return 6
            else:
                raise NotImplementedError(
                    "DiamondLattice.getUniqueLayerCount: Input direction is not supported."
                )
        return ValueError("The input direction is not correct.")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.common.unittest as unittest
import pytest

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.core.expr.visitor import identify_variables
from pyomo.core.expr.compare import compare_expressions
from idaes.apps.nmpc.cost_expressions import (
    get_tracking_cost_from_constant_setpoint,
)


@pytest.mark.unit
class TestTrackingCost(unittest.TestCase):
    def test_tracking_cost_no_weights(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = {
            str(pyo.ComponentUID(m.v1)): 3.0,
            str(pyo.ComponentUID(m.v2)): 4.0,
        }

        m.tracking_expr = get_tracking_cost_from_constant_setpoint(
            [m.v1, m.v2],
            m.time,
            setpoint_data,
        )

        var_sets = {
            i: ComponentSet(identify_variables(m.tracking_expr[i])) for i in m.time
        }
        for i in m.time:
            self.assertIn(m.v1[i], var_sets[i])
            self.assertIn(m.v2[i], var_sets[i])
            pred_value = (1 * i - 3) ** 2 + (2 * i - 4) ** 2
            self.assertEqual(pred_value, pyo.value(m.tracking_expr[i]))
            pred_expr = (m.v1[i] - 3) ** 2 + (m.v2[i] - 4) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.tracking_expr[i].expr))

    def test_tracking_cost_with_weights(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = {
            str(pyo.ComponentUID(m.v1)): 3.0,
            str(pyo.ComponentUID(m.v2)): 4.0,
        }
        weight_data = {
            str(pyo.ComponentUID(m.v1)): 0.1,
            str(pyo.ComponentUID(m.v2)): 0.5,
        }

        m.tracking_expr = get_tracking_cost_from_constant_setpoint(
            [m.v1, m.v2],
            m.time,
            setpoint_data,
            weight_data=weight_data,
        )

        var_sets = {
            i: ComponentSet(identify_variables(m.tracking_expr[i])) for i in m.time
        }
        for i in m.time:
            self.assertIn(m.v1[i], var_sets[i])
            self.assertIn(m.v2[i], var_sets[i])
            pred_value = 0.1 * (1 * i - 3) ** 2 + 0.5 * (2 * i - 4) ** 2
            self.assertAlmostEqual(pred_value, pyo.value(m.tracking_expr[i]))
            pred_expr = 0.1 * (m.v1[i] - 3) ** 2 + 0.5 * (m.v2[i] - 4) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.tracking_expr[i].expr))

    def test_exceptions(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = {
            str(pyo.ComponentUID(m.v1)): 3.0,
        }
        weight_data = {
            str(pyo.ComponentUID(m.v1)): 0.1,
        }

        with self.assertRaisesRegex(KeyError, "Setpoint data"):
            m.tracking_expr = get_tracking_cost_from_constant_setpoint(
                [m.v1, m.v2],
                m.time,
                setpoint_data,
            )

        setpoint_data = {
            str(pyo.ComponentUID(m.v1)): 3.0,
            str(pyo.ComponentUID(m.v2)): 4.0,
        }

        with self.assertRaisesRegex(KeyError, "Tracking weight"):
            m.tracking_expr = get_tracking_cost_from_constant_setpoint(
                [m.v1, m.v2],
                m.time,
                setpoint_data,
                weight_data=weight_data,
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import idaes
import os
import pandas as pd
import numpy as np
from scipy import sparse
import pyomo.contrib.parmest.parmest as parmest
from pyomo.environ import *
from pyomo.opt import SolverFactory
import shutil
import logging
from collections import namedtuple
from idaes.apps.uncertainty_propagation.sens import (
    SensitivityInterface,
    get_dsdp,
    get_dfds_dcds,
)

# will replace with pyomo
# (Pyomo PR 1613: https://github.com/pyomo/pyomo/pull/1613/)

logger = logging.getLogger("idaes.apps.uncertainty_propagation")


def quantify_propagate_uncertainty(
    model_function,
    model_uncertain,
    data,
    theta_names,
    obj_function=None,
    tee=False,
    diagnostic_mode=False,
    solver_options=None,
    covariance_n=None,
):
    """This function calculates error propagation of the objective function and
    constraints. The parmest uses 'model_function' to estimate uncertain
    parameters. The uncertain parameters in 'model_uncertain' are fixed with
    the estimated values. The function 'quantify_propagate_uncertainty'
    calculates error propagation of objective function and constraints
    in the 'model_uncertain'.

    The following terms are used to define the output dimensions:
    Ncon   = number of constraints
    Nvar   = number of variables (Nx + Ntheta)
    Nx     = the number of decision (primal) variables
    Ntheta = number of uncertain parameters.

    Parameters
    ----------
    model_function : function
        A python Function that generates an instance of the Pyomo model using
        'data' as the input argument
    model_uncertain : function or Pyomo ConcreteModel
        Function is a python/ Function that generates an instance of the
        Pyomo model
    data : pandas DataFrame, list of dictionary, or list of json file names
        Data that is used to build an instance of the Pyomo model and build the
        objective function
    theta_names : list of strings
        List of Var names to estimate
    obj_function : function, optional
        Function used to formulate parameter estimation objective, generally
        sum of squared error between measurements and model variables,
        by default None
    tee : bool, optional
        Indicates that ef solver output should be teed, by default False
    diagnostic_mode : bool, optional
        If True, print diagnostics from the solver, by default False
    solver_options : dict, optional
        Provides options to the solver (also the name of an attribute),
        by default None
    covariance_n : int, optional
        Number of datapoints to use in the objective function to
        calculate the covariance matrix.  If omitted, defaults to
        len(data)

    Returns
    -------
    tuple
        results object containing the all information including

        - results.obj: float
            Real number. Objective function value for the given obj_function
        - results.theta: dict
            Size Ntheta python dictionary. Estimated parameters
        - results.theta_names: list
            Size Ntheta list. Names of parameters
        - results.cov: numpy.ndarray
            Ntheta by Ntheta matrix. Covariance of theta
        - results.gradient_f: numpy.ndarray
            Length Nvar array. Gradient vector of the objective function with
            respect to the (decision variables, parameters) at the optimal
            solution
        - results.gradient_c: scipy.sparse.csr.csr_matrix
            Ncon by Nvar size sparse matrix. Gradient vector of the constraints
            with respect to the (decision variables, parameters) at the optimal
            solution.
        - results.dsdp: scipy.sparse.csr.csr_matrix
            Ntheta by Nvar size sparse matrix. Gradient vector of the
            (decision variables, parameters) with respect to parameters
            (=theta_name). number of rows = len(theta_name),
            number of columns= len(col)
        - results.propagation_c: numpy.ndarray
            Length Ncon array. Error propagation in the constraints,
            dc/dp*cov_p*dc/dp + (dc/dx*dx/dp)*cov_p*(dc/dx*dx/dp)
        - results.propagation_f: numpy.float64
            Real number. Error propagation in the objective function,
            df/dp*cov_p*df/dp + (df/dx*dx/dp)*cov_p*(df/dx*dx/dp)
        - results.col: list
            Size Nvar. List of variable names. Note that variables names
            includes both decision variable and uncertain parameter names.
            The order can be mixed.
        - results.row: list
            Size Ncon+1. List of constraints and objective function names

    Raises
    ------
    TypeError
        When tee entry is not Boolean
    TypeError
        When diagnostic_mode entry is not Boolean
    TypeError
        When solver_options entry is not None and a Dictionary
    Warnings
        When an element of theta_names includes a space

    """

    if not isinstance(tee, bool):
        raise TypeError("tee  must be boolean.")
    if not isinstance(diagnostic_mode, bool):
        raise TypeError("diagnostic_mode  must be boolean.")
    if not solver_options == None:
        if not isinstance(solver_options, dict):
            raise TypeError("solver_options must be dictionary.")
    if covariance_n is None:
        if isinstance(data, pd.DataFrame):
            covariance_n = len(data.index)
        else:
            covariance_n = len(data)
        logger.info(
            "covariance_n omitted from quantify_propagate_uncertainty().  "
            f"Assuming {covariance_n}"
        )
    # Remove all "'" and " " in theta_names
    theta_names, var_dic, variable_clean = clean_variable_name(theta_names)
    parmest_class = parmest.Estimator(
        model_function,
        data,
        theta_names,
        obj_function,
        tee,
        diagnostic_mode,
        solver_options,
    )
    obj, theta, cov = parmest_class.theta_est(calc_cov=True, cov_n=covariance_n)
    # Convert theta keys to the original name
    # Revert theta_names to be original
    if variable_clean:
        theta_out = {}
        for i in var_dic.keys():
            theta_out[var_dic[i]] = theta[i]
        theta_names = [var_dic[v] for v in theta_names]
    else:
        theta_out = theta
    propagate_results = propagate_uncertainty(
        model_uncertain, theta, cov, theta_names, tee
    )

    Output = namedtuple(
        "Output",
        [
            "obj",
            "theta",
            "theta_names",
            "cov",
            "gradient_f",
            "gradient_c",
            "dsdp",
            "propagation_c",
            "propagation_f",
            "col",
            "row",
        ],
    )
    results = Output(
        obj,
        theta_out,
        theta_names,
        cov,
        propagate_results.gradient_f,
        propagate_results.gradient_c,
        propagate_results.dsdp,
        propagate_results.propagation_c,
        propagate_results.propagation_f,
        propagate_results.col,
        propagate_results.row,
    )
    return results


def propagate_uncertainty(
    model_uncertain, theta, cov, theta_names, tee=False, solver_options=None
):
    """This function calculates gradient vector, expectation, and variance of
    the objective function and constraints  of the model for given estimated
    optimal parameters and covariance matrix of parameters.
    It calculates error propagation of the objective function and constraints
    by using gradient vector and covariance matrix.

    The following terms are used to define the output dimensions:
    Ncon   = number of constraints
    Nvar   = number of variables (Nx + Ntheta)
    Nx     = the number of decision (primal) variables
    Np = number of uncertain parameters.

    Parameters
    ----------
    model_uncertain : function or Pyomo ConcreteModel
        Function is a python/ Function that generates an instance of the
        Pyomo model
    theta : dict
        Size Ntheta python dictionary. Estimated parameters
    cov : numpy.ndarray
        Ntheta by Ntheta matrix. Covariance matrix of parameters
    theta_names : list of strings
        Size Ntheta. List of estimated l theta names
    tee : bool, optional
        Indicates that ef solver output should be teed, by default False
    solver_options : dict, optional
        Provides options to the solver (also the name of an attribute),
        by default None

    Returns
    -------
    tuple
        results object containing the all information including

        - results.gradient_f: numpy.ndarray
            Length Nvar array. Gradient vector of the objective function with
            respect to the (decision variables, parameters) at the optimal
            solution
        - results.gradient_c: scipy.sparse.csr.csr_matrix
            Ncon by Nvar size sparse matrix. Gradient vector of the constraints
            with respect to the (decision variables, parameters) at the optimal
            solution.
        - results.dsdp: scipy.sparse.csr.csr_matrix
            Ntheta by Nvar size sparse matrix. Gradient vector of the
            (decision variables, parameters) with respect to parameters
            (=theta_name). number of rows = len(theta_name),
            number of columns= len(col)
        - results.propagation_c: numpy.ndarray
            Length Ncon array. Error propagation in the constraints,
            dc/dp*cov_p*dc/dp + (dc/dx*dx/dp)*cov_p*(dc/dx*dx/dp)
        - results.propagation_f: numpy.float64
            Real number. Error propagation in the objective function,
            df/dp*cov_p*df/dp + (df/dx*dx/dp)*cov_p*(df/dx*dx/dp)
        - results.col: list
            Size Nvar. List of variable names. Note that variables names
            includes both decision variable and uncertain parameter names.
            The order can be mixed.
        - results.row: list
            Size Ncon+1. List of constraints and objective function names

    Raises
    ------
    Exception
        if model_uncertain is neither 'ConcreteModel' nor 'function'.
    """
    # define Pyomo model
    try:
        if isinstance(model_uncertain, Block):
            model = model_uncertain
        else:
            model = model_uncertain()
    except TypeError as e:
        raise """
        model_uncertain must be either python function or Pyomo ConcreteModel"""

    # check and convert (optional) covariance matrix
    if isinstance(cov, np.ndarray):
        cov_ = cov
    else:
        cov_ = cov.to_numpy()

    if len(cov_.shape) != 2:
        raise ValueError("cov must be a 2-dimensional matrix or dataframe")

    if cov_.shape[0] != cov_.shape[1]:
        raise ValueError("cov must be square")

    if cov_.shape[0] != len(theta_names):
        raise ValueError(
            """cov must be a n x n matrix or dataframe where 
                         n = len(theta_names)"""
        )

    if len(theta_names) != len(theta):
        raise ValueError(
            """theta_names and theta must have the same number 
                          of elements"""
        )

    # Remove all "'" in theta_names
    theta_names, var_dic, variable_clean = clean_variable_name(theta_names)
    for v in theta_names:
        model.find_component(var_dic[v]).setlb(theta[v])
        model.find_component(var_dic[v]).setub(theta[v])
    # get gradient of the objective function, constraints,
    # and the column,row names
    dsdp, col = get_dsdp(model, theta_names, theta, var_dic, tee)
    dsdp = dsdp.toarray().T  # change shape, Nvar by Ntheta
    gradient_f, gradient_c, col, row, line_dic = get_dfds_dcds(model, theta_names, tee)
    num_constraints = len(
        list(model.component_data_objects(Constraint, active=True, descend_into=True))
    )

    # calculate error propagation of the objective function
    # = df/dp*cov_p*df/dp + (df/dx*dx/dp)*cov_p*(df/dx*dx/dp)
    # = (df/ds*ds/dp)*cov*(df/ds*ds/dp)
    # step 1. df/ds*ds/dp

    # [1 x (Nx + Np)] matrix
    df_ds = np.reshape(gradient_f, (1, len(gradient_f)))

    # [(Nx + Np) x (Np) ] matrix
    ds_dp = np.reshape(dsdp, (len(dsdp), len(theta_names)))

    # [1 x Np ] matrix
    fssp = np.matmul(df_ds, ds_dp)

    # step 2. (df/ds*ds/dp)*cov*(df/ds*ds/dp).transpose()
    # [1 x Np] x [Np x Np ] x [Np x 1] = [1 x 1]
    propagation_f = fssp @ cov_ @ fssp.transpose()
    #
    assert propagation_f.shape == (
        1,
        1,
    ), """propagation_f should be 
    a 1 x 1 matrix. Something is wrong if you are seeing this..."""

    # convert to scalar
    propagation_f = propagation_f[0, 0]

    # calculate error propagation of constraints
    # = dc/dp*cov_p*dc/dp + (dc/dx*dx/dp)*cov_p*(dc/dx*dx/dp)
    # = (dc/ds*ds/dp)*cov*(dc/ds*ds/dp)
    num_constraints = len(
        list(model.component_data_objects(Constraint, active=True, descend_into=True))
    )
    if num_constraints > 0:
        # gradient_c rearrange.
        # k_aug sparse form is [col_idx, row_idx, val] with index starts from 1
        # python sparse form is [row_idx, col_idx, val] with index starts from 0
        # note: variable 'row' from get_dfds_dcds includes objecive function
        # name. i.e, Ncon = len(row)-1
        """
        row_idx = gradient_c[:,1]-1
        col_idx = gradient_c[:,0]-1
        data = gradient_c[:,2]
        gradient_c
        = sparse.csr_matrix((data, (row_idx, col_idx)),
                             shape=(len(row)-1, len(col)))
        """
        # step 1. dc/ds*ds/dp
        # [Ncon x (Nx + Np)] x [(Nx + Np] x Np] = [Ncon x Np] matrix
        cssp = np.matmul(gradient_c.toarray(), dsdp)

        # step 2. (dc/ds*ds/dp)*cov*(dc/ds*ds/dp).transpose()
        # [Ncon x Np] x [Np x Np] x [Np x Ncon]

        # Not sure if this is correct.
        # propagation_c = np.sum(np.matmul(cssp,cov)*cssp,axis=1)

        # Updated to match propgation_f
        propagation_c = cssp @ cov_ @ cssp.transpose()
    else:
        propagation_c = np.array([])

    Output = namedtuple(
        "Output",
        [
            "gradient_f",
            "gradient_c",
            "dsdp",
            "propagation_c",
            "propagation_f",
            "col",
            "row",
        ],
    )
    results = Output(
        gradient_f,
        gradient_c,
        sparse.csr_matrix(dsdp),
        propagation_c,
        propagation_f,
        col,
        row,
    )
    return results


# TODO: Improve the robustness of Parmest then remove this function.
def clean_variable_name(theta_names):
    """This function removes all ' and spaces in theta_names. Note that
    the current theta_est(calc_cov=True) of parmest in Pyomo doesn't allow ' and
    spaces in the variable names. Once a future version of Parmest fixes this issue,
    this function can be depreciated.

    Parameters
    ----------
    theta_names : list of strings
        List of Var names

    Returns
    -------
        It returns the following variables

        - theta_names_out: list of strings
            List of Var names after removing  all ' and spaces
        - var_dic: dict
            Dictionary with keys converted theta_names and values original
            theta_names
    """
    # Variable names cannot have "'" for parmest_class.theta_est(calc_cov=True)
    # Save original variables name in to var_dic
    # Remove all "'" and " " in theta_names
    var_dic = {}
    theta_names_out = []
    clean = False
    for i in range(len(theta_names)):
        theta_tmp = theta_names[i].replace("'", "")
        theta_tmp = theta_tmp.replace(" ", "")
        theta_names_out.append(theta_tmp)
        var_dic[theta_tmp] = theta_names[i]
        if "'" in theta_names[i] or " " in theta_names[i]:
            logger.warning(theta_names[i] + " includes ' or space.")
            logger.warning("The cleaned name: " + theta_tmp)
            clean = True
    if clean:
        logger.warning("All ' and spaces in theta_names are removed.")
    return theta_names_out, var_dic, clean



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pyomo.environ as aml
import pyomo.dae as dae
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core.util.model_statistics import degrees_of_freedom
import pytest

__author__ = "Robert Parker"


def make_model(horizon=5, nfe=10, ncp=2):
    """The simplest DAE I can think of. An isothermal CSTR with one
    reaction. No nlp solver should have trouble with this one.
    Other than the flow_out*conc_out term in the material_balances,
    this model is entirely linear.

    """
    m = aml.ConcreteModel()
    m.time = dae.ContinuousSet(bounds=(0, horizon))
    m.components = aml.Set(initialize=["A", "B"])

    m.volume = aml.Param(initialize=1.0)
    m.max_height = aml.Var(initialize=1.0)
    m.max_height.fix()

    stoich_dict = {
        "A": -1,
        "B": 1,
    }
    m.stoich = aml.Param(m.components, initialize=stoich_dict)
    m.k_rxn = aml.Param(initialize=1.0)
    m.rxn_order = aml.Param(initialize=1, domain=aml.Integers)

    # Variables
    m.flow_in = aml.Var(m.time)
    m.flow_out = aml.Var(m.time)

    m.conc_in = aml.Var(m.time, m.components)
    m.conc = aml.Var(m.time, m.components)

    m.rate = aml.Var(m.time, m.components)

    m.dcdt = dae.DerivativeVar(m.conc, wrt=m.time)

    # Equations
    def mb_rule(m, t, j):
        return (
            m.volume * m.dcdt[t, j]
            == m.flow_in[t] * m.conc_in[t, j]
            - m.flow_out[t] * m.conc[t, j]
            + m.volume * m.rate[t, j]
        )

    m.material_balance = aml.Constraint(m.time, m.components, rule=mb_rule)

    def flow_rule(m, t):
        return m.flow_in[t] == m.flow_out[t]

    m.flow_eqn = aml.Constraint(m.time, rule=flow_rule)

    def rate_rule(m, t, j):
        return m.rate[t, j] == m.stoich[j] * m.k_rxn * m.conc[t, "A"] ** m.rxn_order

    m.rate_eqn = aml.Constraint(m.time, m.components, rule=rate_rule)

    disc = aml.TransformationFactory("dae.collocation")
    disc.apply_to(m, wrt=m.time, nfe=nfe, ncp=2, scheme="LAGRANGE-RADAU")

    # Degrees of freedom:
    m.conc_in[:, "A"].fix(5.0)
    m.conc_in[:, "B"].fix(0.0)

    m.conc[0, "A"].fix(0.0)
    m.conc[0, "B"].fix(0.0)

    m.flow_in.fix(1.0)

    return m


def make_small_model():
    return make_model(horizon=1, nfe=2)


def initialize_t0(model):
    time = model.time
    t0 = time.first()
    calculate_variable_from_constraint(
        model.flow_out[t0],
        model.flow_eqn[t0],
    )
    for j in model.components:
        calculate_variable_from_constraint(
            model.rate[t0, j],
            model.rate_eqn[t0, j],
        )
        calculate_variable_from_constraint(
            model.dcdt[t0, j],
            model.material_balance[t0, j],
        )


def copy_values_forward(model):
    time = model.time
    t0 = time.first()
    for t in time:
        for j in model.components:
            model.conc[t, j].set_value(model.conc[t0, j].value)
            model.rate[t, j].set_value(model.rate[t0, j].value)
            model.dcdt[t, j].set_value(model.dcdt[t0, j].value)
        model.flow_out[t].set_value(model.flow_out[t0].value)


@pytest.mark.unit
def test_model():
    m = make_model()
    assert degrees_of_freedom(m) == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test SolidOxideCell model using parameters from the literature. Compare to both
literature data and cached data from the SolidOxideCell model (to detect model
changes having a significant effect on the model outputs)

This test is based on two papers. The first is Herring et al., in which results
of a series of water electrolysis experiments on a ten cell stack is reported.
Current-voltage curves for the stack are reported at a variety of flow rates,
stream compositions, and temperatures. Some of these results are then
replicated in the CFD software package FLUENT.

The second paper is Kazempoor and Braun, a modeling paper in which some of
these experimental results are replicated, and parameter values for this
replication are reported. These parameter values are used to validate the
SolidOxideCell model.

Several factors limit the ability of the SolidOxideCell model to fully
replicate the Herring et al. data:
    1) The experimental stack operates using cross flow, whereas the
    SolidOxideCell model is limited to counter and co current operation.
    2) The SolidOxideCell model is adiabatic, whereas the stack in Herring et
    al. was operated in a furnace, and received a substantial amount of heat
    3) The open cell potential in Herring et al. was somewhat below the Nernst
    potential, whereas the SolidOxideCell model does not attempt to model
    current leaks
    
However, Kazempoor and Braun use a cocurrent and (probably?) adiabatic model to
partially replicate the Herring et al. experimental data. As a result, the
SolidOxideCell can replicate their replication dataset more closely than the
actual experimental data. Because that model has a slightly different structure
than the SolidOxideCell and the open circuit potential it reports agrees with the
experimental data instead of the Nernst potential calculated both by the thermo
model used by the SolidOxideCell, which agrees with the theoretical open cell
potential reported by Herring et al., there is not a perfect match between
the Kazempoor dataset and that generated by the SolidOxideCell. 

J. Stephen Herring, James E. O’Brien, Carl M. Stoots, G.L. Hawkes, Joseph J.
Hartvigsen, Mehrdad Shahnam, Progress in high-temperature electrolysis for
hydrogen production using planar SOFC technology, International Journal of
Hydrogen Energy, Volume 32, Issue 4, 2007, Pages 440-450, ISSN 0360-3199,
https://doi.org/10.1016/j.ijhydene.2006.06.061.
(https://www.sciencedirect.com/science/article/pii/S0360319906002709)

P. Kazempoor, R.J. Braun, Model validation and performance analysis of
regenerative solid oxide cells: Electrolytic operation, International Journal
of Hydrogen Energy, Volume 39, Issue 6, 2014, Pages 2669-2684, ISSN 0360-3199,
https://doi.org/10.1016/j.ijhydene.2013.12.010.
(https://www.sciencedirect.com/science/article/pii/S036031991302908X)
"""

__author__ = "Douglas Allan"

import os
from copy import deepcopy
import numpy as np
from scipy.interpolate import interp1d
import pytest
import pandas as pd
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir

from idaes.core import FlowsheetBlock
from idaes.models.properties import iapws95
from idaes.models.properties.general_helmholtz import helmholtz_data_dir as hdir
import idaes.core.util.scaling as iscale
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.power_generation.properties.natural_gas_PR import (
    get_prop,
    EosType,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels import (
    SolidOxideCell,
    SolidOxideModuleSimple,
)
from idaes.models_extra.power_generation.unit_models.soc_submodels.common import (
    _comp_enthalpy_expr,
    _comp_entropy_expr,
)
from idaes.models.properties.general_helmholtz import helmholtz_available
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern

data_cache = os.sep.join([this_file_dir(), "data_cache"])


def cccm_to_mps(V):
    # Takes cccm, returns moles per second
    V = V * 1e-6 / 60
    P = 101300
    R = 8.31
    T = 273.15
    return V * P / (R * T)


def mps_to_cccm(n):
    P = 101300
    R = 8.31
    T = 273.15
    V = n * R * T / P
    return V * 1e6 * 60


N_cell = 10
V_min = 8
P = 85910

zfaces = np.linspace(0, 1, 11).tolist()

xfaces_electrode = [0.0, 1.0]
xfaces_electrolyte = [0.0, 1.0]
xfaces_interconnect = [0.0, 1.0]

fuel_comps = ["H2", "H2O", "N2"]
fuel_stoich_dict = {
    "H2": -0.5,
    "H2O": 0.5,
    "N2": 0,
    "Vac": 0.5,
    "O^2-": -0.5,
    "e^-": 1.0,
}
oxygen_comps = ["O2", "N2"]
oxygen_stoich_dict = {"O2": -0.25, "Vac": -0.5, "O^2-": 0.5, "e^-": -1.0}

cell_config = {
    "has_holdup": False,
    "control_volume_zfaces": zfaces,
    "control_volume_xfaces_fuel_electrode": xfaces_electrode,
    "control_volume_xfaces_oxygen_electrode": xfaces_electrode,
    "control_volume_xfaces_electrolyte": xfaces_electrolyte,
    "control_volume_xfaces_interconnect": xfaces_interconnect,
    "fuel_component_list": fuel_comps,
    "fuel_triple_phase_boundary_stoich_dict": fuel_stoich_dict,
    "inert_fuel_species_triple_phase_boundary": ["N2"],
    "oxygen_component_list": oxygen_comps,
    "oxygen_triple_phase_boundary_stoich_dict": oxygen_stoich_dict,
    "inert_oxygen_species_triple_phase_boundary": ["N2"],
    "flow_pattern": HeatExchangerFlowPattern.cocurrent,
    "include_temperature_x_thermo": True,
    "include_contact_resistance": True,
    "flux_through_interconnect": True,
}


def fix_cell_parameters(cell):
    cell.fuel_channel.length_x.fix(1.09e-3)
    cell.oxygen_channel.length_x.fix(1.09e-3)

    cell.length_y.fix(0.08)

    cell.length_z.fix(0.08)
    cell.fuel_channel.heat_transfer_coefficient.fix(100)

    cell.contact_flow_mesh_fuel_electrode.log_preexponential_factor.fix(
        pyo.log(0.55e-4)
    )
    cell.contact_flow_mesh_oxygen_electrode.log_preexponential_factor.fix(
        pyo.log(0.55e-4)
    )
    cell.contact_interconnect_fuel_flow_mesh.log_preexponential_factor.fix(
        pyo.log(0.15e-4)
    )
    cell.contact_interconnect_oxygen_flow_mesh.log_preexponential_factor.fix(
        pyo.log(0.15e-4)
    )

    cell.contact_flow_mesh_fuel_electrode.thermal_exponent_dividend.fix(0)
    cell.contact_flow_mesh_fuel_electrode.contact_fraction.fix(1)
    cell.contact_flow_mesh_oxygen_electrode.thermal_exponent_dividend.fix(0)
    cell.contact_flow_mesh_oxygen_electrode.contact_fraction.fix(1)

    cell.contact_interconnect_fuel_flow_mesh.thermal_exponent_dividend.fix(0)
    cell.contact_interconnect_fuel_flow_mesh.contact_fraction.fix(1)
    cell.contact_interconnect_oxygen_flow_mesh.thermal_exponent_dividend.fix(0)
    cell.contact_interconnect_oxygen_flow_mesh.contact_fraction.fix(1)

    cell.oxygen_channel.heat_transfer_coefficient.fix(100)

    cell.fuel_electrode.length_x.fix(23e-6)
    cell.fuel_electrode.porosity.fix(0.37)
    cell.fuel_electrode.tortuosity.fix(3)
    cell.fuel_electrode.solid_heat_capacity.fix(450)
    cell.fuel_electrode.solid_density.fix(3210.0)
    cell.fuel_electrode.solid_thermal_conductivity.fix(2.16)
    cell.fuel_electrode.resistivity_log_preexponential_factor.fix(pyo.log(8.856e-6))
    cell.fuel_electrode.resistivity_thermal_exponent_dividend.fix(0)

    cell.oxygen_electrode.length_x.fix(21e-6)
    cell.oxygen_electrode.porosity.fix(0.37)
    cell.oxygen_electrode.tortuosity.fix(3.0)
    cell.oxygen_electrode.solid_heat_capacity.fix(430)
    cell.oxygen_electrode.solid_density.fix(3030)
    cell.oxygen_electrode.solid_thermal_conductivity.fix(2.16)
    cell.oxygen_electrode.resistivity_log_preexponential_factor.fix(pyo.log(1.425e-4))
    cell.oxygen_electrode.resistivity_thermal_exponent_dividend.fix(0)

    cell.electrolyte.length_x.fix(1.4e-4)
    cell.electrolyte.heat_capacity.fix(470)
    cell.electrolyte.density.fix(5160)
    cell.electrolyte.thermal_conductivity.fix(2.16)
    cell.electrolyte.resistivity_log_preexponential_factor.fix(pyo.log(1.07e-4))
    cell.electrolyte.resistivity_thermal_exponent_dividend.fix(7237)

    cell.interconnect.length_x.fix(4.6e-4)
    cell.interconnect.heat_capacity.fix(630)
    cell.interconnect.density.fix(7700)
    cell.interconnect.thermal_conductivity.fix(25)
    cell.interconnect.resistivity_log_preexponential_factor.fix(pyo.log(1.2e-6))
    cell.interconnect.resistivity_thermal_exponent_dividend.fix(0)

    # Values from Kazempoor and Braun
    cell.fuel_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(5.5e8)
    )
    cell.fuel_triple_phase_boundary.exchange_current_activation_energy.fix(100e3)
    cell.fuel_triple_phase_boundary.activation_potential_alpha1.fix(1)
    cell.fuel_triple_phase_boundary.activation_potential_alpha2.fix(1)
    cell.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2"].fix(0.5)
    cell.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2O"].fix(0.1)

    cell.oxygen_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(7e8)
    )
    cell.oxygen_triple_phase_boundary.exchange_current_activation_energy.fix(110e3)
    cell.oxygen_triple_phase_boundary.activation_potential_alpha1.fix(1)
    cell.oxygen_triple_phase_boundary.activation_potential_alpha2.fix(1)
    cell.oxygen_triple_phase_boundary.exchange_current_exponent_comp["O2"].fix(0.5)


def model_func():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(
        dynamic=False,
        time_set=[0],
        time_units=pyo.units.s,
    )

    m.fs.propertiesIapws95 = iapws95.Iapws95ParameterBlock()
    m.fs.prop_Iapws95 = iapws95.Iapws95StateBlock(parameters=m.fs.propertiesIapws95)

    m.fs.cell = SolidOxideCell(**cell_config)

    fix_cell_parameters(m.fs.cell)

    m.fs.cell.oxygen_inlet.flow_mol[0].fix(cccm_to_mps(3500) / N_cell)
    m.fs.cell.oxygen_inlet.mole_frac_comp[0, "O2"].fix(0.21)
    m.fs.cell.oxygen_inlet.mole_frac_comp[0, "N2"].fix(0.79)

    m.fs.cell.recursive_scaling()

    return m


@pytest.fixture
def model():
    # Want to be able to call model_func when calling this test directly
    # so that the reference dataset can be recreated if necessary
    return model_func()


@pytest.fixture
def model_stack():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(
        dynamic=False,
        time_set=[0],
        time_units=pyo.units.s,
    )

    m.fs.propertiesIapws95 = iapws95.Iapws95ParameterBlock()
    m.fs.prop_Iapws95 = iapws95.Iapws95StateBlock(parameters=m.fs.propertiesIapws95)

    m.fs.oxygen_params = GenericParameterBlock(
        **get_prop(oxygen_comps, {"Vap"}, eos=EosType.IDEAL),
    )
    m.fs.fuel_params = GenericParameterBlock(
        **get_prop(fuel_comps, {"Vap"}, eos=EosType.IDEAL),
    )

    m.fs.stack = SolidOxideModuleSimple(
        solid_oxide_cell_config=cell_config,
        fuel_property_package=m.fs.fuel_params,
        oxygen_property_package=m.fs.oxygen_params,
    )

    fix_cell_parameters(m.fs.stack.solid_oxide_cell)
    m.fs.stack.number_cells.fix(10)
    m.fs.stack.oxygen_inlet.flow_mol[0].fix(cccm_to_mps(3500))
    m.fs.stack.oxygen_inlet.mole_frac_comp[0, "O2"].fix(0.21)
    m.fs.stack.oxygen_inlet.mole_frac_comp[0, "N2"].fix(0.79)

    for pp in [m.fs.fuel_params, m.fs.oxygen_params]:
        pp.set_default_scaling("flow_mol", 1e3)
        pp.set_default_scaling("flow_mol_phase", 1e3)
        pp.set_default_scaling("mole_frac_comp", 10)
        pp.set_default_scaling("mole_frac_phase_comp", 10)
        pp.set_default_scaling("enth_mol_phase", 1e-3)
        pp.set_default_scaling("pressure", 1e-5)
        pp.set_default_scaling("temperature", 1e-2)

    return m


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_initialization_cell(model):
    m = model
    cell = m.fs.cell

    cell.potential.fix(1.288)
    cell.fuel_inlet.temperature[0].fix(1103.15)
    cell.fuel_inlet.pressure.fix(85910)
    cell.fuel_inlet.flow_mol[0].fix(0.000484863)
    cell.fuel_inlet.mole_frac_comp[0, "H2"].fix(0.0630491)
    cell.fuel_inlet.mole_frac_comp[0, "H2O"].fix(0.6273812)
    cell.fuel_inlet.mole_frac_comp[0, "N2"].fix(0.3095697)

    cell.oxygen_inlet.temperature[0].fix(1103.15)
    cell.oxygen_inlet.pressure.fix(85910)

    cell.recursive_scaling()

    cell.initialize_build(
        optarg={"nlp_scaling_method": "user-scaling"},
        current_density_guess=0,
        temperature_guess=1103.15,
    )
    cell.model_check()
    # Test whether fixed degrees of freedom remain fixed
    assert degrees_of_freedom(m.fs.cell) == 0

    approx = lambda x: pytest.approx(x, 5e-3)
    assert cell.current_density[0, 1].value == approx(-2394.77)
    assert cell.current_density[0, 3].value == approx(-2326.71)
    assert cell.current_density[0, 5].value == approx(-2268.31)
    assert cell.current_density[0, 8].value == approx(-2191.66)
    assert cell.current_density[0, 10].value == approx(-2145.27)

    assert cell.fuel_outlet.temperature[0].value == approx(1103.40)
    assert pyo.value(cell.fuel_outlet.pressure[0]) == approx(85910)
    assert approx(484.863e-6) == pyo.value(cell.fuel_outlet.flow_mol[0])
    assert approx(0.472735) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "H2O"])
    assert approx(0.217695) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "H2"])
    assert approx(0.309570) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "N2"])

    assert cell.oxygen_outlet.temperature[0].value == approx(1103.42)
    assert pyo.value(cell.oxygen_outlet.pressure[0]) == approx(85910)
    assert approx(297.821e-6) == pyo.value(cell.oxygen_outlet.flow_mol[0])
    assert approx(0.3094487) == pyo.value(cell.oxygen_outlet.mole_frac_comp[0, "O2"])
    assert approx(0.690551) == pyo.value(cell.oxygen_outlet.mole_frac_comp[0, "N2"])

    # Test whether unfixed degrees of freedom remain unfixed
    cell.potential.unfix()
    cell.fuel_inlet.temperature[0].unfix()
    cell.fuel_inlet.pressure.unfix()
    cell.fuel_inlet.flow_mol[0].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "H2"].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "H2O"].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "N2"].unfix()

    cell.oxygen_inlet.temperature[0].unfix()
    cell.oxygen_inlet.pressure.unfix()
    cell.oxygen_inlet.mole_frac_comp[0, "O2"].unfix()
    cell.oxygen_inlet.mole_frac_comp[0, "N2"].unfix()

    assert degrees_of_freedom(cell) == 11

    cell.initialize(current_density_guess=0, temperature_guess=1103.15)

    assert degrees_of_freedom(cell) == 11

    # Clean up side effects for other tests
    cell.oxygen_inlet.mole_frac_comp[0, "O2"].fix()
    cell.oxygen_inlet.mole_frac_comp[0, "N2"].fix()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_initialization_stack(model_stack):
    m = model_stack
    stack = m.fs.stack
    cell = m.fs.stack.solid_oxide_cell

    stack.potential_cell[0].fix(1.288)
    stack.fuel_inlet.temperature[0].fix(1103.15)
    stack.fuel_inlet.pressure.fix(85910)
    stack.fuel_inlet.flow_mol[0].fix(0.00484863)
    stack.fuel_inlet.mole_frac_comp[0, "H2"].fix(0.0630491)
    stack.fuel_inlet.mole_frac_comp[0, "H2O"].fix(0.6273812)
    stack.fuel_inlet.mole_frac_comp[0, "N2"].fix(0.3095697)

    stack.oxygen_inlet.temperature[0].fix(1103.15)
    stack.oxygen_inlet.pressure.fix(85910)

    iscale.calculate_scaling_factors(m)

    assert degrees_of_freedom(m.fs.stack) == 0

    stack.initialize_build(
        optarg={"nlp_scaling_method": "user-scaling"},
        current_density_guess=0,
        temperature_guess=1103.15,
    )
    cell.model_check()
    # Test whether fixed degrees of freedom remain fixed
    assert degrees_of_freedom(m.fs.stack) == 0

    approx = lambda x: pytest.approx(x, 5e-3)
    assert cell.current_density[0, 1].value == approx(-2394.77)
    assert cell.current_density[0, 3].value == approx(-2326.71)
    assert cell.current_density[0, 5].value == approx(-2268.31)
    assert cell.current_density[0, 8].value == approx(-2191.66)
    assert cell.current_density[0, 10].value == approx(-2145.27)

    assert stack.fuel_outlet.temperature[0].value == approx(1103.40)
    assert pyo.value(stack.fuel_outlet.pressure[0]) == approx(85910)
    assert approx(484.863e-5) == pyo.value(stack.fuel_outlet.flow_mol[0])
    assert approx(0.472735) == pyo.value(stack.fuel_outlet.mole_frac_comp[0, "H2O"])
    assert approx(0.217695) == pyo.value(stack.fuel_outlet.mole_frac_comp[0, "H2"])
    assert approx(0.309570) == pyo.value(stack.fuel_outlet.mole_frac_comp[0, "N2"])

    assert stack.oxygen_outlet.temperature[0].value == approx(1103.42)
    assert pyo.value(stack.oxygen_outlet.pressure[0]) == approx(85910)
    assert approx(297.821e-5) == pyo.value(stack.oxygen_outlet.flow_mol[0])
    assert approx(0.3094487) == pyo.value(stack.oxygen_outlet.mole_frac_comp[0, "O2"])
    assert approx(0.690551) == pyo.value(stack.oxygen_outlet.mole_frac_comp[0, "N2"])

    # Test whether unfixed degrees of freedom remain unfixed
    stack.potential_cell.unfix()
    stack.fuel_inlet.temperature[0].unfix()
    stack.fuel_inlet.pressure.unfix()
    stack.fuel_inlet.flow_mol[0].unfix()
    stack.fuel_inlet.mole_frac_comp[0, "H2"].unfix()
    stack.fuel_inlet.mole_frac_comp[0, "H2O"].unfix()
    stack.fuel_inlet.mole_frac_comp[0, "N2"].unfix()

    stack.oxygen_inlet.temperature[0].unfix()
    stack.oxygen_inlet.pressure.unfix()
    stack.oxygen_inlet.mole_frac_comp[0, "O2"].unfix()
    stack.oxygen_inlet.mole_frac_comp[0, "N2"].unfix()

    assert degrees_of_freedom(stack) == 11

    stack.initialize_build(
        optarg={"nlp_scaling_method": "user-scaling"},
        current_density_guess=0,
        temperature_guess=1103.15,
    )

    assert degrees_of_freedom(stack) == 11

    # Clean up side effects for other tests
    stack.oxygen_inlet.mole_frac_comp[0, "O2"].fix()
    stack.oxygen_inlet.mole_frac_comp[0, "N2"].fix()


def kazempoor_braun_replication(model):
    m = model
    cell = m.fs.cell
    case = 5

    solver = get_solver("ipopt")

    N_voltage = 11
    N_case = 5
    _constF = 96485
    df = pd.read_csv(os.sep.join([data_cache, "herring-et-al-data.csv"]), index_col=0)

    fuel_out = cell.fuel_outlet
    oxygen_out = cell.oxygen_outlet
    tracked_vars = {
        "voltage": cell.potential[0],
        "average_current_density": sum(cell.current_density[0, j] for j in cell.iznodes)
        / len(cell.iznodes),
        "average_temperature_z": sum(cell.temperature_z[0, j] for j in cell.iznodes)
        / len(cell.iznodes),
        "fuel_out_flow_mol": fuel_out.flow_mol[0],
        "fuel_out_temperature": fuel_out.temperature[0],
        "fuel_out_pressure": fuel_out.pressure[0],
        "fuel_out_mol_frac_comp[H2]": fuel_out.mole_frac_comp[0, "H2"],
        "fuel_out_mol_frac_comp[H2O]": fuel_out.mole_frac_comp[0, "H2O"],
        "fuel_out_mol_frac_comp[N2]": fuel_out.mole_frac_comp[0, "N2"],
        "oxygen_out_flow_mol": oxygen_out.flow_mol[0],
        "oxygen_out_temperature": oxygen_out.temperature[0],
        "oxygen_out_pressure": oxygen_out.pressure[0],
        "oxygen_out_mol_frac_comp[O2]": oxygen_out.mole_frac_comp[0, "O2"],
        "oxygen_out_mol_frac_comp[N2]": oxygen_out.mole_frac_comp[0, "N2"],
    }

    results = dict()
    for key in tracked_vars.keys():
        results[key] = []

    for case in range(1, N_case + 1):
        T_in = df["T_in"][case] + 273.15
        T_dew = df["T_dew"][case] + 273.15
        N_N2 = cccm_to_mps(df["sccm_N2"][case])
        N_H2 = cccm_to_mps(df["sccm_H2"][case])

        # IAPWS95 returns psat in kPa
        p_H2O = 1e3 * pyo.value(
            m.fs.prop_Iapws95.p_sat_func("H2O", 647.096 / T_dew, hdir)
        )
        y_H2O = p_H2O / P
        N_H2O = (N_N2 + N_H2) * y_H2O / (1 - y_H2O)

        N_tot = N_N2 + N_H2 + N_H2O

        y_H2 = N_H2 / N_tot

        assert abs(y_H2O * N_tot - N_H2O) < 1e-15

        dHrxn = (
            _comp_enthalpy_expr(T_in, "H2")
            + 0.5 * _comp_enthalpy_expr(T_in, "O2")
            - _comp_enthalpy_expr(T_in, "H2O")
        )
        dSrxn = (
            _comp_entropy_expr(T_in, "H2")
            + 0.5 * _comp_entropy_expr(T_in, "O2")
            - _comp_entropy_expr(T_in, "H2O")
        )
        dGrxn = dHrxn - T_in * dSrxn

        V_nernst = pyo.value(
            1
            / (2 * _constF)
            * (
                dGrxn
                - 8.31
                * T_in
                * pyo.log(y_H2O / (y_H2 * 0.21**0.5) * (P / 101300) ** -0.5)
            )
        )

        m.fs.cell.potential.fix(V_nernst)
        m.fs.cell.fuel_inlet.temperature[0].fix(T_in)
        m.fs.cell.fuel_inlet.pressure.fix(P)
        m.fs.cell.fuel_inlet.flow_mol[0].fix(N_tot / N_cell)
        m.fs.cell.fuel_inlet.mole_frac_comp[0, "H2"].fix(y_H2)
        m.fs.cell.fuel_inlet.mole_frac_comp[0, "H2O"].fix(y_H2O)
        m.fs.cell.fuel_inlet.mole_frac_comp[0, "N2"].fix(N_N2 / N_tot)

        m.fs.cell.oxygen_inlet.temperature[0].fix(T_in)
        m.fs.cell.oxygen_inlet.pressure.fix(P)

        m.fs.cell.recursive_scaling()

        m.fs.cell.initialize_build(
            optarg={"nlp_scaling_method": "user-scaling"},
            current_density_guess=0,
            temperature_guess=T_in,
        )

        results["voltage"].append(
            np.linspace(V_nernst, 0.9 * df["V_max"][case] / N_cell, N_voltage)
        )
        for key, lst in results.items():
            if key == "voltage":
                continue
            else:
                lst.append([])

        for i in range(N_voltage):
            m.fs.cell.potential.fix(results["voltage"][case - 1][i])
            assert degrees_of_freedom(m) == 0
            status = solver.solve(
                m,
                tee=False,
                options={
                    "tol": 1e-6,
                    "max_iter": 100,
                    "nlp_scaling_method": "user-scaling",
                },
            )
            assert (
                status.solver.termination_condition == pyo.TerminationCondition.optimal
            )
            assert status.solver.status == pyo.SolverStatus.ok
            m.fs.cell.model_check()

            for key, lst in results.items():
                if key == "voltage":
                    continue
                else:
                    lst[case - 1].append(pyo.value(tracked_vars[key]))

    df = pd.DataFrame(results)
    out = []

    for row in df.iterrows():
        tmp_dict = {}
        for key, val in row[1].items():
            tmp_dict[key] = np.array(val)
        tmp_df = pd.DataFrame(tmp_dict)
        out.append(tmp_df)

    return out


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_model_replication(model):
    out = kazempoor_braun_replication(model)
    cached_results = []
    for i in range(len(out)):
        cached_results.append(
            pd.read_csv(
                os.sep.join([data_cache, f"case_{i+1}_interconnect.csv"]), index_col=0
            )
        )

    for df, cached_df in zip(out, cached_results):
        pd.testing.assert_frame_equal(
            df, cached_df, check_dtype=False, check_exact=False, rtol=3e-3
        )

    df = out[4]
    data = pd.read_csv(
        os.sep.join([data_cache, "sweep_5_kazempoor_replication.csv"]),
        names=["average_current_density", "voltage"],
    )
    intr = interp1d(
        x=data["voltage"].to_numpy(), y=data["average_current_density"].to_numpy()
    )
    for row in df.iterrows():
        assert intr(row[1]["voltage"]) == pytest.approx(
            -row[1]["average_current_density"], rel=3e-2, abs=50
        )


def model_func_voltage_drop_custom():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=[0], time_units=pyo.units.s)

    config = deepcopy(cell_config)
    config["voltage_drop_custom"] = True

    m.fs.propertiesIapws95 = iapws95.Iapws95ParameterBlock()
    m.fs.prop_Iapws95 = iapws95.Iapws95StateBlock(parameters=m.fs.propertiesIapws95)

    m.fs.cell = SolidOxideCell(**config)

    fix_cell_parameters(m.fs.cell)

    m.fs.cell.oxygen_inlet.flow_mol[0].fix(cccm_to_mps(3500) / N_cell)
    m.fs.cell.oxygen_inlet.mole_frac_comp[0, "O2"].fix(0.21)
    m.fs.cell.oxygen_inlet.mole_frac_comp[0, "N2"].fix(0.79)

    m.fs.cell.fuel_electrode.voltage_drop_custom.fix(0)
    m.fs.cell.oxygen_electrode.voltage_drop_custom.fix(0)
    m.fs.cell.electrolyte.voltage_drop_custom.fix(0)
    m.fs.cell.interconnect.voltage_drop_custom.fix(0)
    m.fs.cell.fuel_triple_phase_boundary.voltage_drop_custom.fix(0)
    m.fs.cell.oxygen_triple_phase_boundary.voltage_drop_custom.fix(0)

    m.fs.cell.recursive_scaling()

    return m


@pytest.fixture
def model_vdc():
    # Want to be able to call model_func when calling this test directly
    # so that the reference dataset can be recreated if necessary
    return model_func_voltage_drop_custom()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_initialization_cell_voltage_drop_custom(model_vdc):
    m = model_vdc
    cell = m.fs.cell

    cell.potential.fix(1.288)
    cell.fuel_inlet.temperature[0].fix(1103.15)
    cell.fuel_inlet.pressure.fix(85910)
    cell.fuel_inlet.flow_mol[0].fix(0.000484863)
    cell.fuel_inlet.mole_frac_comp[0, "H2"].fix(0.0630491)
    cell.fuel_inlet.mole_frac_comp[0, "H2O"].fix(0.6273812)
    cell.fuel_inlet.mole_frac_comp[0, "N2"].fix(0.3095697)

    cell.oxygen_inlet.temperature[0].fix(1103.15)
    cell.oxygen_inlet.pressure.fix(85910)

    cell.recursive_scaling()

    cell.initialize_build(
        optarg={"nlp_scaling_method": "user-scaling"},
        current_density_guess=0,
        temperature_guess=1103.15,
    )
    cell.model_check()
    # Test whether fixed degrees of freedom remain fixed
    assert degrees_of_freedom(m.fs.cell) == 0

    approx = lambda x: pytest.approx(x, 5e-3)
    assert cell.current_density[0, 1].value == approx(-2394.77)
    assert cell.current_density[0, 3].value == approx(-2326.71)
    assert cell.current_density[0, 5].value == approx(-2268.31)
    assert cell.current_density[0, 8].value == approx(-2191.66)
    assert cell.current_density[0, 10].value == approx(-2145.27)

    assert cell.fuel_outlet.temperature[0].value == approx(1103.40)
    assert pyo.value(cell.fuel_outlet.pressure[0]) == approx(85910)
    assert approx(484.863e-6) == pyo.value(cell.fuel_outlet.flow_mol[0])
    assert approx(0.472735) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "H2O"])
    assert approx(0.217695) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "H2"])
    assert approx(0.309570) == pyo.value(cell.fuel_outlet.mole_frac_comp[0, "N2"])

    assert cell.oxygen_outlet.temperature[0].value == approx(1103.42)
    assert pyo.value(cell.oxygen_outlet.pressure[0]) == approx(85910)
    assert approx(297.821e-6) == pyo.value(cell.oxygen_outlet.flow_mol[0])
    assert approx(0.3094487) == pyo.value(cell.oxygen_outlet.mole_frac_comp[0, "O2"])
    assert approx(0.690551) == pyo.value(cell.oxygen_outlet.mole_frac_comp[0, "N2"])

    for iz in cell.iznodes:
        # H2O is consumed at fuel electrode, concentration at electrode surface
        # should be less than that in channel, and concentration at TPB should
        # be less than that at the surface
        assert pyo.value(cell.fuel_channel.conc_mol_comp_deviation_x1[0, iz, "H2O"]) < 0
        assert (
            pyo.value(
                cell.fuel_electrode.conc_mol_comp_deviation_x1[0, iz, "H2O"]
                - cell.fuel_channel.conc_mol_comp_deviation_x1[0, iz, "H2O"]
            )
            < 0
        )
        # H2 is produced at fuel electrode, concentration at electrode surface
        # should be greater than that in channel, and concentration at TPB should
        # be greater than that at the surface
        assert pyo.value(cell.fuel_channel.conc_mol_comp_deviation_x1[0, iz, "H2"]) > 0
        assert (
            pyo.value(
                cell.fuel_electrode.conc_mol_comp_deviation_x1[0, iz, "H2"]
                - cell.fuel_channel.conc_mol_comp_deviation_x1[0, iz, "H2"]
            )
            > 0
        )
        # O2 is produced at oxygen electrode, concentration at electrode surface
        # should be greater than that in channel, and concentration at TPB should
        # be greater than that at the surface
        assert (
            pyo.value(cell.oxygen_channel.conc_mol_comp_deviation_x0[0, iz, "O2"]) > 0
        )
        assert (
            pyo.value(
                cell.oxygen_electrode.conc_mol_comp_deviation_x0[0, iz, "O2"]
                - cell.oxygen_channel.conc_mol_comp_deviation_x0[0, iz, "O2"]
            )
            > 0
        )

    # Test whether unfixed degrees of freedom remain unfixed
    cell.potential.unfix()
    cell.fuel_inlet.temperature[0].unfix()
    cell.fuel_inlet.pressure.unfix()
    cell.fuel_inlet.flow_mol[0].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "H2"].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "H2O"].unfix()
    cell.fuel_inlet.mole_frac_comp[0, "N2"].unfix()

    cell.oxygen_inlet.temperature[0].unfix()
    cell.oxygen_inlet.pressure.unfix()
    cell.oxygen_inlet.mole_frac_comp[0, "O2"].unfix()
    cell.oxygen_inlet.mole_frac_comp[0, "N2"].unfix()

    assert degrees_of_freedom(cell) == 11

    cell.initialize(current_density_guess=0, temperature_guess=1103.15)

    assert degrees_of_freedom(cell) == 11

    # Clean up side effects for other tests
    cell.oxygen_inlet.mole_frac_comp[0, "O2"].fix()
    cell.oxygen_inlet.mole_frac_comp[0, "N2"].fix()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_model_replication_voltage_drop_custom(model_vdc):
    out = kazempoor_braun_replication(model_vdc)
    cached_results = []
    for i in range(len(out)):
        cached_results.append(
            pd.read_csv(
                os.sep.join([data_cache, f"case_{i+1}_interconnect.csv"]), index_col=0
            )
        )

    for df, cached_df in zip(out, cached_results):
        pd.testing.assert_frame_equal(
            df, cached_df, check_dtype=False, check_exact=False, rtol=3e-3
        )

    df = out[4]
    data = pd.read_csv(
        os.sep.join([data_cache, "sweep_5_kazempoor_replication.csv"]),
        names=["average_current_density", "voltage"],
    )
    intr = interp1d(
        x=data["voltage"].to_numpy(), y=data["average_current_density"].to_numpy()
    )
    for row in df.iterrows():
        assert intr(row[1]["voltage"]) == pytest.approx(
            -row[1]["average_current_density"], rel=3e-2, abs=50
        )


if __name__ == "__main__":
    m = model_func()
    out = kazempoor_braun_replication(m)
    # Uncomment to recreate cached data
    for i, df in enumerate(out):
        df.to_csv(os.sep.join([data_cache, f"case_{i+1}_interconnect.csv"]))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""
import pytest
from pyomo.environ import check_optimal_termination, ConcreteModel, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)


from idaes.models_extra.column_models.properties.MEA_vapor import flue_gas, wet_co2


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestFlueGasBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**flue_gas)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].calculate_scaling_factors()

        # Fix state
        model.props[1].flow_mol.fix(21.48)
        model.props[1].temperature.fix(317.88)
        model.props[1].pressure.fix(107650)
        model.props[1].mole_frac_comp["CO2"].fix(0.11453)
        model.props[1].mole_frac_comp["H2O"].fix(0.08526)
        model.props[1].mole_frac_comp["N2"].fix(0.73821)
        model.props[1].mole_frac_comp["O2"].fix(0.06200)

        return model

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Values verified with older model
        assert pytest.approx(30.6689467, rel=1e-8) == value(
            model.props[1].cp_mol_phase["Vap"]
        )

        assert pytest.approx(1.77021548e-05, rel=1e-8) == value(
            model.props[1].visc_d_phase["Vap"]
        )

        assert pytest.approx(2.50126426e-2, rel=1e-8) == value(
            model.props[1].therm_cond_phase["Vap"]
        )

        assert pytest.approx(1.75238546e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "CO2"]
        )
        assert pytest.approx(2.64401213e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "H2O"]
        )
        assert pytest.approx(2.06616153e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "N2"]
        )
        assert pytest.approx(2.14369756e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "O2"]
        )


class TestWetCO2Block(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**wet_co2)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].calculate_scaling_factors()

        # Fix state
        model.props[1].flow_mol.fix(17.496)
        model.props[1].temperature.fix(396.6)
        model.props[1].pressure.fix(183430)
        model.props[1].mole_frac_comp["CO2"].fix(0.0145)
        model.props[1].mole_frac_comp["H2O"].fix(0.9855)

        return model

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Values verified with older model
        assert pytest.approx(34.3944239, rel=1e-8) == value(
            model.props[1].cp_mol_phase["Vap"]
        )

        assert pytest.approx(1.36007669e-05, rel=1e-8) == value(
            model.props[1].visc_d_phase["Vap"]
        )

        assert pytest.approx(2.65105253e-2, rel=1e-8) == value(
            model.props[1].therm_cond_phase["Vap"]
        )

        assert pytest.approx(1.90464661e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "CO2"]
        )
        assert pytest.approx(1.90464661e-05, rel=1e-8) == value(
            model.props[1].diffus_phase_comp["Vap", "H2O"]
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test Wilke mixing rules for viscosity from Properties of Gases and Liquids, 5th Ed.
Section 9-5-2

Authors: Douglas Allan
"""

import pytest
from sys import modules

from pyomo.environ import ConcreteModel, value, Var, units as pyunits
from pyomo.util.check_units import assert_units_equivalent
import pyomo.environ as pyo

from idaes.models.properties.modular_properties.transport_properties.viscosity_wilke import (
    ViscosityWilke,
    wilke_phi_ij_callback,
    herring_zimmer_phi_ij_callback,
)

from idaes.core import declare_process_block_class, LiquidPhase, VaporPhase

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.pure import ConstantProperties


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def construct_dummy_model(component_dict, callback):
    m = ConcreteModel()
    components = [comp for comp in component_dict.keys()]
    comp1 = components[0]
    comp2 = components[1]

    m.params = DummyParameterBlock(
        components={
            comp1: {
                "visc_d_phase_comp": {"Vap": ConstantProperties},
                "parameter_data": {
                    "mw": (component_dict[comp1]["mw"], pyunits.g / pyunits.mol),
                },
            },
            comp2: {
                "visc_d_phase_comp": {"Vap": ConstantProperties},
                "parameter_data": {
                    "mw": (component_dict[comp2]["mw"], pyunits.g / pyunits.mol),
                },
            },
        },
        phases={
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Ideal,
                "visc_d_phase": ViscosityWilke,
                "transport_property_options": {"viscosity_phi_ij_callback": callback},
            },
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)
    ViscosityWilke.visc_d_phase.build_parameters(m.params.Vap)

    # Add common variables
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )
    m.props[1]._visc_d_phase_comp = Var(
        m.props[1].phase_list,
        m.props[1].component_list,
        initialize=0,
        units=pyunits.Pa * pyunits.s,
    )
    m.props[1]._visc_d_phase_comp["Vap", comp1].value = pyunits.convert(
        component_dict[comp1]["visc_d_Vap"] * pyunits.micropoise,
        to_units=pyunits.Pa * pyunits.s,
    )
    m.props[1]._visc_d_phase_comp["Vap", comp2].value = pyunits.convert(
        component_dict[comp2]["visc_d_Vap"] * pyunits.micropoise,
        to_units=pyunits.Pa * pyunits.s,
    )
    return m


@pytest.mark.unit
def test_wilke_visc_d_phase_N2_CO2():
    component_dict = {
        "N2": {"mw": 28.014, "visc_d_Vap": 175.8},
        "CO2": {"mw": 44.009, "visc_d_Vap": 146.6},
    }
    m = construct_dummy_model(component_dict, wilke_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    phi_N2_N2 = m.props[1].visc_d_phi_ij["N2", "N2"]
    assert_units_equivalent(phi_N2_N2, pyunits.dimensionless)
    assert pyo.value(phi_N2_N2) == pytest.approx(1, rel=1e-12)

    phi_CO2_CO2 = m.props[1].visc_d_phi_ij["CO2", "CO2"]
    assert_units_equivalent(phi_CO2_CO2, pyunits.dimensionless)
    assert pyo.value(phi_CO2_CO2) == pytest.approx(1, rel=1e-12)

    phi_N2_CO2 = m.props[1].visc_d_phi_ij["N2", "CO2"]
    assert_units_equivalent(phi_N2_CO2, pyunits.dimensionless)
    assert pyo.value(phi_N2_CO2) == pytest.approx(1.369409683998896, rel=1e-12)

    phi_CO2_N2 = m.props[1].visc_d_phi_ij["CO2", "N2"]
    assert_units_equivalent(phi_CO2_N2, pyunits.dimensionless)
    # Eq. 9-5.15, Properties of Gases and Liquids 5th Ed.
    assert pyo.value(phi_CO2_N2) == pytest.approx(
        pyo.value(phi_N2_CO2 * 146.6 / 175.8 * 28.014 / 44.009), rel=1e-12
    )

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.213, 0.495, 0.767, 1.0]  # Mole fraction of N2
    visc_list = [
        146.6,
        153.5,
        161.8,
        172.1,
        175.8,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -1.3, -1.8, -2.8, 0.0]  # Percent error in Wilke's method
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "N2"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "CO2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(err_list[i], abs=0.3)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_wilke_visc_d_phase_ammonia_hydrogen():
    component_dict = {
        "NH3": {"mw": 17.031, "visc_d_Vap": 105.9},
        "H2": {"mw": 2.016, "visc_d_Vap": 90.6},
    }
    m = construct_dummy_model(component_dict, wilke_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.195, 0.399, 0.536, 0.677, 1.0]  # Mole fraction of NH3
    visc_list = [
        90.6,
        118.4,
        123.8,
        122.4,
        120.0,
        105.9,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -11.0, -12.0, -11.0, -9.7, 0.0]  # Percent error in Wilke's method
    for i in range(6):
        m.props[1].mole_frac_phase_comp["Vap", "NH3"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "H2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(err_list[i], abs=0.3)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
@pytest.mark.unit
def test_wilke_visc_d_phase_N2O_SO2():
    component_dict = {
        "N2O": {"mw": 44.013, "visc_d_Vap": 173.0},
        "SO2": {"mw": 64.066, "visc_d_Vap": 152.3},
    }
    m = construct_dummy_model(component_dict, wilke_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.325, 0.625, 0.817, 1.0]  # Mole fraction of NH3
    visc_list = [
        152.3,
        161.7,
        167.8,
        170.7,
        173.0,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -2.2, -2.2, -1.3, 0.0]  # Percent error in Wilke's method
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "N2O"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "SO2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(err_list[i], abs=0.3)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_herning_zipperer_visc_d_phase_N2_CO2():
    component_dict = {
        "N2": {"mw": 28.014, "visc_d_Vap": 175.8},
        "CO2": {"mw": 44.009, "visc_d_Vap": 146.6},
    }
    m = construct_dummy_model(component_dict, herring_zimmer_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    phi_N2_N2 = m.props[1].visc_d_phi_ij["N2", "N2"]
    assert_units_equivalent(phi_N2_N2, pyunits.dimensionless)
    assert pyo.value(phi_N2_N2) == pytest.approx(1, rel=1e-12)

    phi_CO2_CO2 = m.props[1].visc_d_phi_ij["CO2", "CO2"]
    assert_units_equivalent(phi_CO2_CO2, pyunits.dimensionless)
    assert pyo.value(phi_CO2_CO2) == pytest.approx(1, rel=1e-12)

    phi_N2_CO2 = m.props[1].visc_d_phi_ij["N2", "CO2"]
    assert_units_equivalent(phi_N2_CO2, pyunits.dimensionless)
    assert pyo.value(phi_N2_CO2) == pytest.approx(1.253381233999109, rel=1e-12)

    phi_CO2_N2 = m.props[1].visc_d_phi_ij["CO2", "N2"]
    assert_units_equivalent(phi_CO2_N2, pyunits.dimensionless)
    # Eq. 9-5.15, Properties of Gases and Liquids 5th Ed.
    assert pyo.value(phi_CO2_N2) == pytest.approx(1 / pyo.value(phi_N2_CO2), rel=1e-12)

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.213, 0.495, 0.767, 1.0]  # Mole fraction of N2
    visc_list = [
        146.6,
        153.5,
        161.8,
        172.1,
        175.8,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -1, -1.5, -2.5, 0.0]  # Percent error in H&Z's method
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "N2"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "CO2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(err_list[i], abs=0.3)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_herning_zipperer_visc_d_phase_ammonia_hydrogen():
    component_dict = {
        "NH3": {"mw": 17.031, "visc_d_Vap": 105.9},
        "H2": {"mw": 2.016, "visc_d_Vap": 90.6},
    }
    m = construct_dummy_model(component_dict, herring_zimmer_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.195, 0.399, 0.536, 0.677, 1.0]  # Mole fraction of NH3
    visc_list = [
        90.6,
        118.4,
        123.8,
        122.4,
        120.0,
        105.9,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -18.0, -19.0, -16.0, -14.0, 0.0]  # Percent error in H&Z's method
    for i in range(6):
        m.props[1].mole_frac_phase_comp["Vap", "NH3"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "H2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(
            err_list[i], abs=1
        )  # Not as many sig figs in the table

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
@pytest.mark.unit
def test_herning_zipperer_visc_d_phase_N2O_SO2():
    component_dict = {
        "N2O": {"mw": 44.013, "visc_d_Vap": 173.0},
        "SO2": {"mw": 64.066, "visc_d_Vap": 152.3},
    }
    m = construct_dummy_model(component_dict, herring_zimmer_phi_ij_callback)

    expr = ViscosityWilke.visc_d_phase.return_expression(m.props[1], "Vap")

    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    mole_frac_list = [0.0, 0.325, 0.625, 0.817, 1.0]  # Mole fraction of NH3
    visc_list = [
        152.3,
        161.7,
        167.8,
        170.7,
        173.0,
    ]  # Experimental viscosities in millipoise
    err_list = [0.0, -2.2, -2.1, -1.2, 0.0]  # Percent error in H&Z's method
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "N2O"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "SO2"] = 1 - mole_frac_list[i]
        err = 100 * value(expr_micropoise / visc_list[i] - 1)
        assert err == pytest.approx(err_list[i], abs=0.3)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ONNXSurrogate
"""
import pytest

pytest.importorskip("onnx", reason="onnx not available")
pytest.importorskip("omlt", reason="omlt not available")

import os.path
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager
from pyomo.environ import (
    ConcreteModel,
    Var,
    SolverFactory,
    assert_optimal_termination,
    value,
)
from idaes.core.surrogate.onnx_surrogate import ONNXSurrogate
from idaes.core.surrogate.surrogate_block import SurrogateBlock
from idaes.core.surrogate.sampling.scaling import OffsetScaler
import json
import onnx

# onnx, onnx_available = attempt_import("onnx")
rtol = 1e-4
atol = 1e-4


def load_onnx_model_data(
    name="net_Calcite_ST",
):
    onnx_folder_name = os.path.join(this_file_dir(), "data", "onnx_models")
    onnx_model = onnx.load(os.path.join(onnx_folder_name, "{}.onnx".format(name)))
    with open(os.path.join(onnx_folder_name, "{}_idaes_info.json".format(name))) as fd:
        scaler_info = json.load(fd)

    test_inputs = {
        "vars": [
            "feed_pH",
            "pressure_bar_feed",
            "Na",
            "Cl",
            "Ca",
            "Mg",
            "HCO3",
            "SO4",
            "K",
            "Sr",
            "Ba",
            "HCl",
        ],
        "feed_pH": 9.5,
        "pressure_bar_feed": 1.01325,
        "Na": 0.230858556000748,
        "Cl": 0.106701648328453,
        "Ca": 0.0245274696499109,
        "Mg": 0.0311348703689873,
        "HCO3": 0.430482141673564,
        "SO4": 0.182204065,
        "K": 0.000500561,
        "Sr": 0.000761853,
        "Ba": 2.50e-05,
        "HCl": 10,
    }
    test_outputs = {"vars": ["Calcite_ST"], "Calcite_ST": 40.4270772546529}
    return onnx_model, scaler_info, test_inputs, test_outputs


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_onnx_surrogate_manual_creation():
    ###
    # Test 1->2 sigmoid
    ###
    onnx_model, scaler_info, test_inputs, test_outputs = load_onnx_model_data()
    input_scaler = None
    for key, items in scaler_info.items():
        print(key, items)
    if scaler_info["input_scaler"] is not None:
        input_scaler = OffsetScaler.from_dict(scaler_info["input_scaler"])

    output_scaler = None
    if scaler_info["output_scaler"] is not None:
        output_scaler = OffsetScaler.from_dict(scaler_info["output_scaler"])
    onnx_surrogate = ONNXSurrogate(
        onnx_model,
        input_labels=scaler_info["input_labels"],
        output_labels=scaler_info["output_labels"],
        input_bounds=scaler_info["input_bounds"],
        input_scaler=input_scaler,
        output_scaler=output_scaler,
    )

    m = ConcreteModel()

    output_vars = ["Calcite_ST"]
    m.inputs = Var(test_inputs["vars"])

    m.outputs = Var(test_outputs["vars"])
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=onnx_surrogate,
        input_vars=[m.inputs[input_var] for input_var in test_inputs["vars"]],
        output_vars=[m.outputs[output_var] for output_var in test_outputs["vars"]],
        formulation=ONNXSurrogate.Formulation.REDUCED_SPACE,
    )
    for key in test_inputs["vars"]:
        m.inputs[key].fix(test_inputs[key])

    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)
    assert pytest.approx(test_outputs["Calcite_ST"], rel=1e-3) == value(
        m.outputs["Calcite_ST"]
    )


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_onnx_surrogate_load_and_save_from_file():
    ###
    # Test 1->2 sigmoid
    ###
    _, _, test_inputs, test_outputs = load_onnx_model_data()

    onnx_surrogate = ONNXSurrogate.load_onnx_model(
        onnx_model_location=os.path.join(this_file_dir(), "data", "onnx_models"),
        model_name="net_Calcite_ST",
    )
    with TempfileManager.new_context() as tf:
        dname = tf.mkdtemp()
        onnx_surrogate.save_to_folder(dname, "temp_model")

        loaded_onnx_surrogate = ONNXSurrogate.load_onnx_model(
            onnx_model_location=dname,
            model_name="temp_model",
        )
    assert not os.path.isdir(dname)
    m = ConcreteModel()

    output_vars = ["Calcite_ST"]
    m.inputs = Var(test_inputs["vars"])

    m.outputs = Var(test_outputs["vars"])
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=loaded_onnx_surrogate,
        input_vars=[m.inputs[input_var] for input_var in test_inputs["vars"]],
        output_vars=[m.outputs[output_var] for output_var in test_outputs["vars"]],
        formulation=ONNXSurrogate.Formulation.REDUCED_SPACE,
    )
    for key in test_inputs["vars"]:
        m.inputs[key].fix(test_inputs[key])

    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)
    assert pytest.approx(test_outputs["Calcite_ST"], rel=1e-3) == value(
        m.outputs["Calcite_ST"]
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Wassiljew-Mason-Saxena mixing rules for thermal conductivity.
"""

import pytest
from sys import modules

from pyomo.environ import ConcreteModel, value, Var, units as pyunits
from pyomo.util.check_units import assert_units_equivalent
import pyomo.environ as pyo

from idaes.models.properties.modular_properties.transport_properties import (
    ThermalConductivityWMS,
)
from idaes.models.properties.modular_properties.transport_properties.viscosity_wilke import (
    wilke_phi_ij_callback,
)
from idaes.core import declare_process_block_class, LiquidPhase, VaporPhase

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.pure import (
    ConstantProperties,
    ChapmanEnskog,
)


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def construct_dummy_model(component_dict, chapman_enskog=False):
    m = ConcreteModel()
    components = [comp for comp in component_dict.keys()]
    comp1 = components[0]
    comp2 = components[1]

    config_dict = {
        comp1: {
            "therm_cond_phase_comp": {"Vap": ConstantProperties.Constant},
            "parameter_data": {
                "mw": (component_dict[comp1]["mw"], pyunits.g / pyunits.mol),
                "therm_cond_Vap_comp_coeff": (
                    component_dict[comp1]["therm_cond_Vap"],
                    pyunits.W / pyunits.m / pyunits.K,
                ),
            },
        },
        comp2: {
            "therm_cond_phase_comp": {"Vap": ConstantProperties.Constant},
            "parameter_data": {
                "mw": (component_dict[comp2]["mw"], pyunits.g / pyunits.mol),
                "therm_cond_Vap_comp_coeff": (
                    component_dict[comp2]["therm_cond_Vap"],
                    pyunits.W / pyunits.m / pyunits.K,
                ),
            },
        },
    }
    if chapman_enskog:
        config_dict[comp1]["visc_d_phase_comp"] = {
            "Vap": ChapmanEnskog.ChapmanEnskogLennardJones
        }
        config_dict[comp2]["visc_d_phase_comp"] = {
            "Vap": ChapmanEnskog.ChapmanEnskogLennardJones
        }
        config_dict[comp1]["parameter_data"]["lennard_jones_sigma"] = (
            component_dict[comp1]["sigma"],
            pyunits.angstrom,
        )
        config_dict[comp2]["parameter_data"]["lennard_jones_sigma"] = (
            component_dict[comp2]["sigma"],
            pyunits.angstrom,
        )
        config_dict[comp1]["parameter_data"]["lennard_jones_epsilon_reduced"] = (
            component_dict[comp1]["epsilon"],
            pyunits.K,
        )
        config_dict[comp2]["parameter_data"]["lennard_jones_epsilon_reduced"] = (
            component_dict[comp2]["epsilon"],
            pyunits.K,
        )
    else:
        config_dict[comp1]["visc_d_phase_comp"] = {"Vap": ConstantProperties.Constant}
        config_dict[comp2]["visc_d_phase_comp"] = {"Vap": ConstantProperties.Constant}
        config_dict[comp1]["parameter_data"]["visc_d_Vap_comp_coeff"] = (
            component_dict[comp1]["visc_d_Vap"],
            pyunits.micropoise,
        )
        config_dict[comp2]["parameter_data"]["visc_d_Vap_comp_coeff"] = (
            component_dict[comp2]["visc_d_Vap"],
            pyunits.micropoise,
        )

    m.params = DummyParameterBlock(
        components=config_dict,
        phases={
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Ideal,
                "therm_cond_phase": ThermalConductivityWMS,
            },
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)
    ThermalConductivityWMS.therm_cond_phase.build_parameters(m.params.Vap)
    # Add common variables
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )

    def rule_visc_d_phase_comp(b, p, j):
        cobj = b.params.get_component(j)
        if (
            cobj.config.visc_d_phase_comp is not None
            and p in cobj.config.visc_d_phase_comp
        ):
            return cobj.config.visc_d_phase_comp[p].visc_d_phase_comp.return_expression(
                b, cobj, p, b.temperature
            )
        else:
            return pyo.Expression.Skip

    m.props[1]._visc_d_phase_comp = pyo.Expression(
        m.props[1].phase_list,
        m.props[1].component_list,
        doc="Viscosity for each phase-component pair",
        rule=rule_visc_d_phase_comp,
    )

    m.props[1]._therm_cond_phase_comp = Var(
        m.props[1].phase_list,
        m.props[1].component_list,
        initialize=0,
        units=pyunits.W / pyunits.m / pyunits.K,
    )
    m.props[1]._therm_cond_phase_comp["Vap", comp1].value = component_dict[comp1][
        "therm_cond_Vap"
    ]
    m.props[1]._therm_cond_phase_comp["Vap", comp2].value = component_dict[comp2][
        "therm_cond_Vap"
    ]

    return m


@pytest.mark.unit
def test_wms_therm_cond_phase_():
    # Taken from Example 10-5 of the properties of gases and liquids
    component_dict = {
        "benzene": {"mw": 78.114, "visc_d_Vap": 92.5, "therm_cond_Vap": 1.66e-2},
        "Ar": {"mw": 39.948, "visc_d_Vap": 271, "therm_cond_Vap": 2.14e-2},
    }
    m = construct_dummy_model(component_dict, chapman_enskog=False)

    assert (
        m.params.Vap.config.transport_property_options["viscosity_phi_ij_callback"]
        is wilke_phi_ij_callback
    )

    expr = ThermalConductivityWMS.therm_cond_phase.return_expression(m.props[1], "Vap")
    m.props[1].mole_frac_phase_comp["Vap", "benzene"].value = 0.25
    m.props[1].mole_frac_phase_comp["Vap", "Ar"].value = 0.75

    phi_bb = m.props[1].visc_d_phi_ij["benzene", "benzene"]
    assert_units_equivalent(phi_bb, pyunits.dimensionless)
    assert pyo.value(phi_bb) == pytest.approx(1, rel=1e-12)

    phi_aa = m.props[1].visc_d_phi_ij["Ar", "Ar"]
    assert_units_equivalent(phi_aa, pyunits.dimensionless)
    assert pyo.value(phi_aa) == pytest.approx(1, rel=1e-12)

    phi_ba = m.props[1].visc_d_phi_ij["benzene", "Ar"]
    assert_units_equivalent(phi_ba, pyunits.dimensionless)
    assert pyo.value(phi_ba) == pytest.approx(0.459, rel=2e-3)

    phi_ab = m.props[1].visc_d_phi_ij["Ar", "benzene"]
    assert_units_equivalent(phi_ab, pyunits.dimensionless)
    # Eq. 9-5.15, Properties of Gases and Liquids 5th Ed.
    assert pyo.value(phi_ab) == pytest.approx(2.630, rel=2e-3)

    assert pyo.value(expr) == pytest.approx(1.84e-2, rel=2e-3)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_wms_therm_cond_phase_benzene_n_hexane():
    # Unfortunately, in order to test the thermal conductivity mixing rule, we need to know component
    # viscosities. In order to estimate the visocisity at the particular experimental conditions
    # we use Chapman Enskog with viscosity Lennard Jones parameters from
    # Properties of Gases and Liquids, 5th Ed., Appendix B
    component_dict = {
        "benzene": {
            "mw": 78.114,
            "sigma": 5.349,
            "epsilon": 412.3,
            "therm_cond_Vap": 1.90093e-2,
        },
        "n-hexane": {
            "mw": 86.178,
            "sigma": 5.949,
            "epsilon": 399.3,
            "therm_cond_Vap": 2.29209e-2,
        },
    }
    m = construct_dummy_model(component_dict, chapman_enskog=True)

    assert (
        m.params.Vap.config.transport_property_options["viscosity_phi_ij_callback"]
        is wilke_phi_ij_callback
    )

    expr = ThermalConductivityWMS.therm_cond_phase.return_expression(m.props[1], "Vap")
    # Pulled off Figure 10-6 from Properties of Gases and Liquids, 5th Ed.
    # That table cites: Bennett, L. A., and R. G. Vines: J. Chem. Phys., 23: 1587 (1955).
    mole_frac_list = [0.0, 0.25188, 0.50090, 0.74994, 1.0]  # Mole fraction of benzene
    therm_cond_list = [
        2.29209e-2,
        2.186596e-2,
        2.087267e-2,
        1.997177e-2,
        1.900933e-2,
    ]  # Experimental tcond in W/(m K)
    # The temperature is mislabeled in Figure 10-6. This is the actual temperature from Bennett and Vines (1955)
    m.props[1].temperature.value = 125.0 + 273.15
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "benzene"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "n-hexane"] = 1 - mole_frac_list[i]
        assert value(expr) == pytest.approx(therm_cond_list[i], rel=0.01)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_wms_therm_cond_phase_methanol_n_hexane():
    # Unfortunately, in order to test the thermal conductivity mixing rule, we need to know component
    # viscosities. In order to estimate the visocisity at the particular experimental conditions
    # we use Chapman Enskog with viscosity Lennard Jones parameters from
    # Properties of Gases and Liquids, 5th Ed., Appendix B
    component_dict = {
        "methanol": {
            "mw": 32.04,
            "sigma": 3.626,
            "epsilon": 481.8,
            "therm_cond_Vap": 2.178133e-2,
        },
        "n-hexane": {
            "mw": 86.178,
            "sigma": 5.949,
            "epsilon": 399.3,
            "therm_cond_Vap": 1.98716e-2,
        },
    }
    m = construct_dummy_model(component_dict, chapman_enskog=True)

    assert (
        m.params.Vap.config.transport_property_options["viscosity_phi_ij_callback"]
        is wilke_phi_ij_callback
    )

    expr = ThermalConductivityWMS.therm_cond_phase.return_expression(m.props[1], "Vap")
    # Pulled off Figure 10-6 from Properties of Gases and Liquids, 5th Ed.
    # That table cites: Bennett, L. A., and R. G. Vines:J. Chem. Phys., 23: 1587 (1955).
    mole_frac_list = [
        0.0,
        0.25021,
        0.501179,
        0.748935,
        1.0,
    ]  # Mole fraction of methanol
    therm_cond_list = [
        1.98716e-2,
        2.12654e-2,
        2.22279e-2,
        2.22279e-2,
        2.17813e-2,
    ]  # Experimental tcond in W/(m K)
    m.props[1].temperature.value = 98.4 + 273.15
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "methanol"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "n-hexane"] = 1 - mole_frac_list[i]
        # This mixing rule doesn't appear to be particularly reliable for polar-nonpolar solutions
        assert value(expr) == pytest.approx(therm_cond_list[i], rel=0.08)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)


@pytest.mark.unit
def test_wms_therm_cond_phase_benzene_argon():
    # Unfortunately, in order to test the thermal conductivity mixing rule, we need to know component
    # viscosities. In order to estimate the visocisity at the particular experimental conditions
    # we use Chapman Enskog with viscosity Lennard Jones parameters from
    # Properties of Gases and Liquids, 5th Ed., Appendix B
    component_dict = {
        "benzene": {
            "mw": 78.114,
            "sigma": 5.349,
            "epsilon": 412.3,
            "therm_cond_Vap": 1.663763e-2,
        },
        "Ar": {
            "mw": 39.95,
            "sigma": 3.542,
            "epsilon": 93.3,
            "therm_cond_Vap": 2.13809e-2,
        },
    }
    m = construct_dummy_model(component_dict, chapman_enskog=True)

    assert (
        m.params.Vap.config.transport_property_options["viscosity_phi_ij_callback"]
        is wilke_phi_ij_callback
    )

    expr = ThermalConductivityWMS.therm_cond_phase.return_expression(m.props[1], "Vap")
    # Pulled off Figure 10-6 from Properties of Gases and Liquids, 5th Ed.
    # That table cites: Bennett, L. A., and R. G. Vines:J. Chem. Phys., 23: 1587 (1955).
    mole_frac_list = [0.0, 0.25130, 0.50026, 0.74780, 1.0]  # Mole fraction of benzene
    therm_cond_list = [
        2.13809e-2,
        1.907851e-2,
        1.776181e-2,
        1.709196e-2,
        1.663763e-2,
    ]  # Experimental tcond in W/(m K)
    m.props[1].temperature.value = 100.6 + 273.15
    for i in range(5):
        m.props[1].mole_frac_phase_comp["Vap", "benzene"] = mole_frac_list[i]
        m.props[1].mole_frac_phase_comp["Vap", "Ar"] = 1 - mole_frac_list[i]
        assert value(expr) == pytest.approx(therm_cond_list[i], rel=0.05)

    assert_units_equivalent(expr, pyunits.W / pyunits.m / pyunits.K)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base model for solid-liquid separations.

This model is intended to form the basis for solid-liquid separations where a mixed
stream enters the unit and is separated into a concentrated solid-liquid stream and
a stream of pure liquid (e.g. filters, thickeners, etc.)

This model assumes separate property packages and ports for the solid and liquid
streams and thus two inlet Ports (solid and liquid) and three outlet Ports (solids,
liquid with solids, separated liquids).

"""
# Import Python libraries
import logging
from pandas import DataFrame

# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Port

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.models.unit_models.separator import (
    Separator,
    SplittingType,
    EnergySplittingType,
)
from idaes.core.initialization import BlockTriangularizationInitializer
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.units_of_measurement import report_quantity


__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger("idaes.unit_model")


@declare_process_block_class("SLSeparator")
class SLSeparatorData(UnitModelBlockData):
    """
    Standard Solid-Liquid Separator Unit Model Class
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Flash units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Flash units do not have defined volume, thus
this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "energy_split_basis",
        ConfigValue(
            default=EnergySplittingType.equal_temperature,
            domain=EnergySplittingType,
            description="Type of constraint to write for energy splitting",
            doc="""Argument indicating basis to use for splitting energy this is
not used for when ideal_separation == True.
**default** - EnergySplittingType.equal_temperature.
**Valid values:** {
**EnergySplittingType.equal_temperature** - outlet temperatures equal inlet
**EnergySplittingType.equal_molar_enthalpy** - outlet molar enthalpies equal
inlet,
**EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
flows.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for solid phase",
            doc="""Property parameter object used to define solid phase property
calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing solid phase property packages",
            doc="""A ConfigBlock with arguments to be passed to a solid phase property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for liquid phase",
            doc="""Property parameter object used to define liquid phase property
    calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing liquid phase property packages",
            doc="""A ConfigBlock with arguments to be passed to a liquid phase property
    block(s) and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    default_initializer = BlockTriangularizationInitializer

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()

        # Build Solid Phase - inlet=outlet so only need a StateBlock
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.solid_property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.solid_state = self.config.solid_property_package.build_state_block(
            self.flowsheet().time, doc="Solid properties in separator", **tmp_dict
        )

        # Add Solids Ports
        self.add_port(
            name="solid_inlet", block=self.solid_state, doc="Solid inlet to separator"
        )
        self.add_port(
            name="solid_outlet",
            block=self.solid_state,
            doc="Solid outlet from separator",
        )

        # Add Liquid Inlet
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.liquid_property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.liquid_inlet_state = self.config.liquid_property_package.build_state_block(
            self.flowsheet().time,
            doc="Liquid properties at inlet to separator",
            **tmp_dict,
        )

        self.split = Separator(
            property_package=self.config.liquid_property_package,
            property_package_args=self.config.liquid_property_package_args,
            outlet_list=["recovered", "retained"],
            split_basis=SplittingType.totalFlow,
            ideal_separation=False,
            mixed_state_block=self.liquid_inlet_state,
            has_phase_equilibrium=False,
            material_balance_type=self.config.material_balance_type,
            momentum_balance_type=self.config.momentum_balance_type,
            energy_split_basis=self.config.energy_split_basis,
        )

        # Add liquid ports
        self.add_port(
            name="liquid_inlet",
            block=self.liquid_inlet_state,
            doc="Liquid inlet to separator",
        )
        self.recovered_liquid_outlet = Port(extends=self.split.recovered)
        self.retained_liquid_outlet = Port(extends=self.split.retained)

        # Add liquid recovery
        self.liquid_recovery = Reference(self.split.split_fraction[:, "recovered"])

    def initialize(self, **kwargs):
        raise NotImplementedError(
            "The SLSeparator unit model does not support the old initialization API. "
            "Please use the new API (InitializerObjects) instead."
        )

    def _get_performance_contents(self, time_point=0):

        return {"vars": {"Liquid Recovery": self.liquid_recovery[time_point]}}

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}
        stream_attributes["Units"] = {}

        sblocks = {
            "Solid Inlet": self.solid_state,
            "Liquid Inlet": self.liquid_inlet_state,
            "Solid Outlet": self.solid_state,
            "Liquid in Solids Outlet": self.split.retained_state,
            "Recovered Liquid Outlet": self.split.recovered_state,
        }

        for n, v in sblocks.items():
            dvars = v[time_point].define_display_vars()

            stream_attributes[n] = {}

            for k in dvars:
                for i in dvars[k].keys():
                    stream_key = k if i is None else f"{k} {i}"

                    quant = report_quantity(dvars[k][i])

                    stream_attributes[n][stream_key] = quant.m
                    stream_attributes["Units"][stream_key] = quant.u

        return DataFrame.from_dict(stream_attributes, orient="columns")
